import numpy as np
import pytest

import tidyms2.chem.formula_generator as fg
from tidyms2.chem import Formula, FormulaGeneratorConfiguration, PeriodicTable
from tidyms2.chem.atoms import IsotopeError


@pytest.fixture
def chnops_bounds() -> fg.FormulaCoefficientBounds:
    chnops_str = ["C", "H", "N", "O", "P", "S"]
    bounds = {x: (0, 5) for x in chnops_str}
    return fg.FormulaCoefficientBounds.from_isotope_str(bounds)


def test__FormulaCoefficientBounds_from_isotope_str_negative_bounds_error():
    chnops_str = ["C", "H", "N", "O", "P", "S"]
    bounds = {x: (-1, 5) for x in chnops_str}
    with pytest.raises(ValueError):
        fg.FormulaCoefficientBounds.from_isotope_str(bounds)


def test__FormulaCoefficientBounds_from_isotope_str_invalid_isotope():
    # check that an ValueError is raised if the lower bound is greater than the
    # upper bound
    isotopes = ["14C", "H", "N", "O", "P", "S"]
    bounds = {x: (0, 5) for x in isotopes}
    with pytest.raises(IsotopeError):
        fg.FormulaCoefficientBounds.from_isotope_str(bounds)


def test__FormulaCoefficientBounds_from_isotope_str_invalid_element():
    # check that an ValueError is raised if the lower bound is greater than the
    # upper bound
    isotopes = ["X", "H", "N", "O", "P", "S"]
    bounds = {x: (0, 5) for x in isotopes}
    with pytest.raises(IsotopeError):
        fg.FormulaCoefficientBounds.from_isotope_str(bounds)


def test__FormulaCoefficientBounds_bound_from_mass(chnops_bounds):
    M = 100.0
    tol = 0.005

    # check bounds for 32S. 32 * 4 == 128 > M, 32 * 3 == 96 < M. the upper bound
    # should be 3.
    s32 = PeriodicTable().get_isotope("32S")
    bounds = chnops_bounds.bounds_from_mass(M, tol)
    assert bounds[s32].lower == 0
    assert bounds[s32].upper == 3

    # check bounds for 1HS. The mass-based bound is ~100 >> 5. The result should
    # be 5
    h1 = PeriodicTable().get_isotope("1H")
    assert bounds[h1].lower == 0
    assert bounds[h1].upper == 5


def test__FormulaCoefficientBounds_bound_negative_positive_defect_only_pos_d():
    # check results when using only isotopes with positive mass defect
    isotopes = ["H", "N"]
    bounds = {x: (0, 5) for x in isotopes}
    bounds = fg.FormulaCoefficientBounds.from_isotope_str(bounds)
    d = 0.01
    tol = 0.005
    defect_bounds = bounds.bound_negative_positive_defect(d, tol)
    assert np.isclose(defect_bounds.neg.lower, -tol)
    assert np.isclose(defect_bounds.neg.upper, tol)
    assert np.isclose(defect_bounds.pos.lower, d - tol)
    assert np.isclose(defect_bounds.pos.upper, d + tol)


def test__FormulaCoefficientBounds_bound_negative_positive_defect_only_neg_d():
    # check results when using only isotopes with positive mass defect
    isotopes = ["Cl", "O"]
    bounds = {x: (0, 5) for x in isotopes}
    bounds = fg.FormulaCoefficientBounds.from_isotope_str(bounds)
    d = -0.01
    tol = 0.005
    defect_bounds = bounds.bound_negative_positive_defect(d, tol)
    assert np.isclose(defect_bounds.neg.lower, d - tol)
    assert np.isclose(defect_bounds.neg.upper, d + tol)
    assert np.isclose(defect_bounds.pos.lower, -tol)
    assert np.isclose(defect_bounds.pos.upper, tol)


def test__FormulaCoefficientBounds_get_nominal_defect_candidates(chnops_bounds):
    M = 100.01
    tol = 0.005
    m_candidates, d_candidates = chnops_bounds.get_nominal_defect_candidates(M, tol)
    assert len(m_candidates) == 1
    assert m_candidates[0] == 100
    assert np.isclose(d_candidates[0], 0.01)


def test__FormulaCoefficientBounds_get_nominal_defect_candidates_d_gt_one():
    # check nominal mass and mass defect candidates with mass defect greater
    # than one
    isotopes = ["H"]
    bounds = {x: (190, 200) for x in isotopes}
    bounds = fg.FormulaCoefficientBounds.from_isotope_str(bounds)
    M = 195.5
    tol = 0.001
    m_candidates, d_candidates = bounds.get_nominal_defect_candidates(M, tol)
    assert len(m_candidates) == 1
    assert m_candidates[0] == 194
    assert np.isclose(d_candidates[0], 1.5)


def test__FormulaCoefficientBounds_split_pos_neg(chnops_bounds):
    pos, neg = chnops_bounds.split_pos_neg()
    for x in pos.bounds:
        assert x.d > 0.0

    for x in neg.bounds:
        assert x.d < 0.0


def test__FormulaCoefficientBounds_split_pos_neg_no_positive_d():
    isotopes = ["Cl", "O"]
    bounds = {x: (0, 5) for x in isotopes}
    bounds = fg.FormulaCoefficientBounds.from_isotope_str(bounds)
    pos, neg = bounds.split_pos_neg()
    assert len(pos.bounds) == 0

    for x in neg.bounds:
        assert x.d < 0.0


def test__FormulaCoefficientBounds_split_pos_neg_no_negative_d():
    isotopes = ["H", "N"]
    bounds = {x: (0, 5) for x in isotopes}
    bounds = fg.FormulaCoefficientBounds.from_isotope_str(bounds)
    pos, neg = bounds.split_pos_neg()
    assert len(neg.bounds) == 0

    for x in pos.bounds:
        assert x.d > 0.0


@pytest.mark.parametrize("isotope", ["H", "O"])
def test__FormulaCoefficients_one_isotope_empty(isotope):
    bounds = {x: (0, 0) for x in [isotope]}
    bounds = fg.FormulaCoefficientBounds.from_isotope_str(bounds)
    max_mass = 1000
    return_sorted = False
    return_reversed = False
    coeff = fg.FormulaCoefficients(bounds, max_mass, return_sorted, return_reversed)
    assert coeff.coefficients.shape == (1, 1)
    assert coeff.M.shape == (1,)
    assert coeff.q.shape == (1,)
    assert len(coeff.r_to_index) == 12
    assert len(coeff.r_to_d) == 12


@pytest.mark.parametrize("isotope", ["H", "O"])
def test__FormulaCoefficients_one_isotope(isotope):
    n = 10
    bounds = {x: (0, n) for x in [isotope]}
    bounds = fg.FormulaCoefficientBounds.from_isotope_str(bounds)
    max_mass = 1000
    return_sorted = False
    return_reversed = False
    coeff = fg.FormulaCoefficients(bounds, max_mass, return_sorted, return_reversed)
    assert coeff.coefficients.shape == (n + 1, 1)
    assert coeff.M.shape == (n + 1,)
    assert coeff.q.shape == (n + 1,)
    assert len(coeff.r_to_index) == 12
    assert len(coeff.r_to_d) == 12


@pytest.mark.parametrize("isotope", ["H", "O"])
def test__FormulaCoefficients_one_isotope_return_sorted(isotope):
    n = 10
    bounds = {x: (0, n) for x in [isotope]}
    bounds = fg.FormulaCoefficientBounds.from_isotope_str(bounds)
    max_mass = 1000
    return_sorted = True
    return_reversed = False
    coeff = fg.FormulaCoefficients(bounds, max_mass, return_sorted, return_reversed)
    assert coeff.coefficients.shape == (n + 1, 1)
    assert coeff.M.shape == (n + 1,)
    assert coeff.q.shape == (n + 1,)
    assert len(coeff.r_to_index) == 12
    assert len(coeff.r_to_d) == 12
    for k in range(12):
        assert coeff.r_to_d[k].size == coeff.r_to_index[k].size
        # sort check for mass defect
        diff = np.diff(coeff.r_to_d[k])
        assert (diff >= 0).all()


@pytest.mark.parametrize("isotope", ["H", "O"])
def test__FormulaCoefficients_one_isotope_return_sorted_return_reversed(isotope):
    n = 10
    bounds = {x: (0, n) for x in [isotope]}
    bounds = fg.FormulaCoefficientBounds.from_isotope_str(bounds)
    max_mass = 1000
    return_sorted = True
    return_reversed = True
    coeff = fg.FormulaCoefficients(bounds, max_mass, return_sorted, return_reversed)
    assert coeff.coefficients.shape == (n + 1, 1)
    assert coeff.M.shape == (n + 1,)
    assert coeff.q.shape == (n + 1,)
    assert len(coeff.r_to_index) == 12
    assert len(coeff.r_to_d) == 12
    for k in range(12):
        assert coeff.r_to_d[k].size == coeff.r_to_index[k].size
        # sort check
        diff = np.diff(coeff.r_to_d[k])
        assert (diff <= 0).all()


@pytest.mark.parametrize("isotopes", [["H", "N"], ["O", "P"]])
def test__FormulaCoefficients_multiple_isotopes(isotopes):
    n = 10
    n_isotopes = len(isotopes)
    n_comb = (n + 1) ** n_isotopes
    bounds = {x: (0, n) for x in isotopes}
    bounds = fg.FormulaCoefficientBounds.from_isotope_str(bounds)
    max_mass = 1000
    return_sorted = False
    return_reversed = False
    coeff = fg.FormulaCoefficients(bounds, max_mass, return_sorted, return_reversed)
    assert coeff.coefficients.shape == (n_comb, n_isotopes)
    assert coeff.M.shape == (n_comb,)
    assert coeff.q.shape == (n_comb,)
    assert len(coeff.r_to_index) == 12
    assert len(coeff.r_to_d) == 12


@pytest.mark.parametrize("isotopes", [["H", "N"], ["O", "P"]])
def test__FormulaCoefficients_multiple_isotopes_return_sorted(isotopes):
    n = 10
    n_isotopes = len(isotopes)
    n_comb = (n + 1) ** n_isotopes
    bounds = {x: (0, n) for x in isotopes}
    bounds = fg.FormulaCoefficientBounds.from_isotope_str(bounds)
    max_mass = 1000
    return_sorted = True
    return_reversed = False
    coeff = fg.FormulaCoefficients(bounds, max_mass, return_sorted, return_reversed)
    assert coeff.coefficients.shape == (n_comb, n_isotopes)
    assert coeff.M.shape == (n_comb,)
    assert coeff.q.shape == (n_comb,)
    assert len(coeff.r_to_index) == 12
    assert len(coeff.r_to_d) == 12
    for k in range(12):
        assert coeff.r_to_d[k].size == coeff.r_to_index[k].size
        # sort check
        diff = np.diff(coeff.r_to_d[k])
        assert (diff >= 0).all()


@pytest.mark.parametrize("isotopes", [["H", "N"], ["O", "P"]])
def test__FormulaCoefficients_multiple_isotopes_return_reversed(isotopes):
    n = 10
    n_isotopes = len(isotopes)
    n_comb = (n + 1) ** n_isotopes
    bounds = {x: (0, n) for x in isotopes}
    bounds = fg.FormulaCoefficientBounds.from_isotope_str(bounds)
    max_mass = 1000
    return_sorted = True
    return_reversed = True
    coeff = fg.FormulaCoefficients(bounds, max_mass, return_sorted, return_reversed)
    assert coeff.coefficients.shape == (n_comb, n_isotopes)
    assert coeff.M.shape == (n_comb,)
    assert coeff.q.shape == (n_comb,)
    assert len(coeff.r_to_index) == 12
    assert len(coeff.r_to_d) == 12
    for k in range(12):
        assert coeff.r_to_d[k].size == coeff.r_to_index[k].size
        # sort check
        diff = np.diff(coeff.r_to_d[k])
        assert (diff <= 0).all()


@pytest.mark.parametrize("formula_str", ["C10H10O2", "C9H14", "C7H8N2"])
def test_FormulaGenerator_brute_force(formula_str):
    # Test correctness of the algorithm by comparing the results with the
    # brute-force solution.
    formula = Formula(formula_str)
    M = formula.get_exact_mass()
    # parameters
    isotope_bounds = {"C": (7, 10), "H": (8, 14), "O": (0, 2), "N": (0, 2)}
    max_mass = 1000
    tol = 0.005
    # brute force solution
    # compute al formula coefficients, sort coefficients by exact mass
    # and search the minimum and maximum valid mass using bisection search.
    bounds = fg.FormulaCoefficientBounds.from_isotope_str(isotope_bounds)
    bf_coeff = bounds.make_coefficients(max_mass)
    bf_isotopes = bf_coeff.isotopes
    bf_coeff_M = bf_coeff.M
    bf_sorted_M_index = np.argsort(bf_coeff_M)
    bf_coeff = bf_coeff.coefficients[bf_sorted_M_index]
    bf_coeff_M = bf_coeff_M[bf_sorted_M_index]
    start_valid, end_valid = np.searchsorted(bf_coeff_M, [M - tol, M + tol])
    bf_coeff = bf_coeff[start_valid:end_valid, :]
    # sort columns by isotope str
    col_sort = np.argsort([str(x) for x in bf_isotopes])
    bf_coeff = bf_coeff[:, col_sort]

    # mass defect based solution
    config = FormulaGeneratorConfiguration(bounds=isotope_bounds, max_M=max_mass)
    f = fg.FormulaGenerator(config)
    f.generate_formulas(M, tol)
    coeff, isotopes, coeff_M = f.results_to_array()
    coeff = coeff[np.argsort(coeff_M), :]
    # sort columns by isotope str
    col_sort = np.argsort([str(x) for x in isotopes])
    coeff = coeff[:, col_sort]
    assert np.array_equal(coeff, bf_coeff)


def test_FormulaGenerator_small_molecules_with_negative_mass_defect():
    config = FormulaGeneratorConfiguration(
        bounds={
            "C": (0, 3),
            "H": (0, 10),
            "N": (0, 4),
            "S": (0, 1),
        },
        max_M=500,
    )

    generator = fg.FormulaGenerator(config)

    tol = 0.005
    f = Formula("S")
    M = f.get_exact_mass() - tol / 2
    generator.generate_formulas(M, tol)
    assert generator.get_n_results() == 1


@pytest.mark.parametrize("f_str", ["H2", "Li"])
def test_FormulaGenerator_small_molecules_with_mass_lower_than_C(f_str: str):
    config = FormulaGeneratorConfiguration(
        bounds={
            "C": (0, 3),
            "H": (0, 10),
            "Li": (0, 1),
        },
        max_M=500,
    )

    generator = fg.FormulaGenerator(config)

    tol = 0.005
    M = Formula("H2").get_exact_mass()
    generator.generate_formulas(M, tol)
    assert generator.get_n_results() == 1


@pytest.mark.parametrize("formula_str", ["C9H14", "C7H8N2"])
def test_FormulaGenerator_brute_force_no_negative_isotopes(formula_str):
    # Test correctness of the algorithm by comparing the results with the
    # brute-force solution.
    formula = Formula(formula_str)
    M = formula.get_exact_mass()
    # parameters
    isotope_bounds = {"C": (7, 10), "H": (8, 14), "N": (0, 2)}
    max_mass = 1000
    tol = 0.005
    # brute force solution
    # compute al formula coefficients, sort coefficients by exact mass
    # and search the minimum and maximum valid mass using bisection search.
    bounds = fg.FormulaCoefficientBounds.from_isotope_str(isotope_bounds)
    bf_coeff = bounds.make_coefficients(max_mass)
    bf_isotopes = bf_coeff.isotopes
    bf_coeff_M = bf_coeff.M
    bf_sorted_M_index = np.argsort(bf_coeff_M)
    bf_coeff = bf_coeff.coefficients[bf_sorted_M_index]
    bf_coeff_M = bf_coeff_M[bf_sorted_M_index]
    start_valid, end_valid = np.searchsorted(bf_coeff_M, [M - tol, M + tol])
    bf_coeff = bf_coeff[start_valid:end_valid, :]
    # sort columns by isotope str
    col_sort = np.argsort([str(x) for x in bf_isotopes])
    bf_coeff = bf_coeff[:, col_sort]

    # mass defect based solution
    config = FormulaGeneratorConfiguration(bounds=isotope_bounds, max_M=max_mass)
    f = fg.FormulaGenerator(config)
    f.generate_formulas(M, tol)
    coeff, isotopes, coeff_M = f.results_to_array()
    coeff = coeff[np.argsort(coeff_M), :]
    # sort columns by isotope str
    col_sort = np.argsort([str(x) for x in isotopes])
    coeff = coeff[:, col_sort]
    assert np.array_equal(coeff, bf_coeff)


@pytest.mark.parametrize("formula_str", ["C9O6", "C7P4"])
def test_FormulaGenerator_brute_force_no_positive_isotopes(formula_str):
    # Test correctness of the algorithm by comparing the results with the
    # brute-force solution.
    formula = Formula(formula_str)
    M = formula.get_exact_mass()
    # parameters
    isotope_bounds = {"C": (7, 10), "O": (0, 10), "P": (0, 5)}
    max_mass = 1000
    tol = 0.005
    # brute force solution
    # compute al formula coefficients, sort coefficients by exact mass
    # and search the minimum and maximum valid mass using bisection search.
    bounds = fg.FormulaCoefficientBounds.from_isotope_str(isotope_bounds)
    bf_coeff = bounds.make_coefficients(max_mass)
    bf_isotopes = bf_coeff.isotopes
    bf_coeff_M = bf_coeff.M
    bf_sorted_M_index = np.argsort(bf_coeff_M)
    bf_coeff = bf_coeff.coefficients[bf_sorted_M_index]
    bf_coeff_M = bf_coeff_M[bf_sorted_M_index]
    start_valid, end_valid = np.searchsorted(bf_coeff_M, [M - tol, M + tol])
    bf_coeff = bf_coeff[start_valid:end_valid, :]
    # sort columns by isotope str
    col_sort = np.argsort([str(x) for x in bf_isotopes])
    bf_coeff = bf_coeff[:, col_sort]

    # mass defect based solution
    config = FormulaGeneratorConfiguration(bounds=isotope_bounds, max_M=max_mass)
    f = fg.FormulaGenerator(config)
    f.generate_formulas(M, tol)
    coeff, isotopes, coeff_M = f.results_to_array()
    coeff = coeff[np.argsort(coeff_M), :]
    # sort columns by isotope str
    col_sort = np.argsort([str(x) for x in isotopes])
    coeff = coeff[:, col_sort]
    assert np.array_equal(coeff, bf_coeff)


@pytest.mark.parametrize("formula_str", ["H2O", "(13C)O2"])
def test_FormulaGenerator_brute_force_no_c12(formula_str):
    # Test correctness of the algorithm by comparing the results with the
    # brute-force solution.
    formula = Formula(formula_str)
    M = formula.get_exact_mass()
    # parameters
    isotope_bounds = {"13C": (0, 5), "O": (0, 5), "H": (0, 5)}
    max_mass = 1000
    tol = 0.005
    # brute force solution
    # compute al formula coefficients, sort coefficients by exact mass
    # and search the minimum and maximum valid mass using bisection search.
    bounds = fg.FormulaCoefficientBounds.from_isotope_str(isotope_bounds)
    bf_coeff = bounds.make_coefficients(max_mass)
    bf_isotopes = bf_coeff.isotopes
    bf_coeff_M = bf_coeff.M
    bf_sorted_M_index = np.argsort(bf_coeff_M)
    bf_coeff = bf_coeff.coefficients[bf_sorted_M_index]
    bf_coeff_M = bf_coeff_M[bf_sorted_M_index]
    start_valid, end_valid = np.searchsorted(bf_coeff_M, [M - tol, M + tol])
    bf_coeff = bf_coeff[start_valid:end_valid, :]
    # sort columns by isotope str
    col_sort = np.argsort([str(x) for x in bf_isotopes])
    bf_coeff = bf_coeff[:, col_sort]

    # mass defect based solution
    config = FormulaGeneratorConfiguration(bounds=isotope_bounds, max_M=max_mass)
    f = fg.FormulaGenerator(config)
    f.generate_formulas(M, tol)
    coeff, isotopes, coeff_M = f.results_to_array()
    coeff = coeff[np.argsort(coeff_M), :]
    # sort columns by isotope str
    col_sort = np.argsort([str(x) for x in isotopes])
    coeff = coeff[:, col_sort]
    assert np.array_equal(coeff, bf_coeff)


@pytest.mark.parametrize("formula_str", ["H3N", "(13C)H4"])
def test_FormulaGenerator_brute_force_no_c12_no_negative(formula_str):
    # Test correctness of the algorithm by comparing the results with the
    # brute-force solution.
    formula = Formula(formula_str)
    M = formula.get_exact_mass()
    # parameters
    isotope_bounds = {"13C": (0, 5), "N": (0, 5), "H": (0, 5)}
    max_mass = 1000
    tol = 0.005
    # brute force solution
    # compute al formula coefficients, sort coefficients by exact mass
    # and search the minimum and maximum valid mass using bisection search.
    bounds = fg.FormulaCoefficientBounds.from_isotope_str(isotope_bounds)
    bf_coeff = bounds.make_coefficients(max_mass)
    bf_isotopes = bf_coeff.isotopes
    bf_coeff_M = bf_coeff.M
    bf_sorted_M_index = np.argsort(bf_coeff_M)
    bf_coeff = bf_coeff.coefficients[bf_sorted_M_index]
    bf_coeff_M = bf_coeff_M[bf_sorted_M_index]
    start_valid, end_valid = np.searchsorted(bf_coeff_M, [M - tol, M + tol])
    bf_coeff = bf_coeff[start_valid:end_valid, :]
    # sort columns by isotope str
    col_sort = np.argsort([str(x) for x in bf_isotopes])
    bf_coeff = bf_coeff[:, col_sort]

    # mass defect based solution
    config = FormulaGeneratorConfiguration(bounds=isotope_bounds, max_M=max_mass)
    f = fg.FormulaGenerator(config)
    f.generate_formulas(M, tol)
    coeff, isotopes, coeff_M = f.results_to_array()
    coeff = coeff[np.argsort(coeff_M), :]
    # sort columns by isotope str
    col_sort = np.argsort([str(x) for x in isotopes])
    coeff = coeff[:, col_sort]
    assert np.array_equal(coeff, bf_coeff)


@pytest.mark.parametrize("formula_str", ["ClO2", "SO3"])
def test_FormulaGenerator_brute_force_no_c12_no_positive(formula_str):
    # Test correctness of the algorithm by comparing the results with the
    # brute-force solution.
    formula = Formula(formula_str)
    M = formula.get_exact_mass()
    # parameters
    isotope_bounds = {"Cl": (0, 5), "O": (0, 5), "S": (0, 5)}
    max_mass = 1000
    tol = 0.005
    # brute force solution
    # compute al formula coefficients, sort coefficients by exact mass
    # and search the minimum and maximum valid mass using bisection search.
    bounds = fg.FormulaCoefficientBounds.from_isotope_str(isotope_bounds)
    bf_coeff = bounds.make_coefficients(max_mass)
    bf_isotopes = bf_coeff.isotopes
    bf_coeff_M = bf_coeff.M
    bf_sorted_M_index = np.argsort(bf_coeff_M)
    bf_coeff = bf_coeff.coefficients[bf_sorted_M_index]
    bf_coeff_M = bf_coeff_M[bf_sorted_M_index]
    start_valid, end_valid = np.searchsorted(bf_coeff_M, [M - tol, M + tol])
    bf_coeff = bf_coeff[start_valid:end_valid, :]
    # sort columns by isotope str
    col_sort = np.argsort([str(x) for x in bf_isotopes])
    bf_coeff = bf_coeff[:, col_sort]

    # mass defect based solution
    config = FormulaGeneratorConfiguration(bounds=isotope_bounds, max_M=max_mass)
    f = fg.FormulaGenerator(config)
    f.generate_formulas(M, tol)
    coeff, isotopes, coeff_M = f.results_to_array()
    coeff = coeff[np.argsort(coeff_M), :]
    # sort columns by isotope str
    col_sort = np.argsort([str(x) for x in isotopes])
    coeff = coeff[:, col_sort]
    assert np.array_equal(coeff, bf_coeff)


def test_FormulaGenerator_generate_formulas_invalid_mass():
    config = FormulaGeneratorConfiguration.from_chnops(500)
    formula_generator = fg.FormulaGenerator(config)
    with pytest.raises(ValueError):
        formula_generator.generate_formulas(-100, 0.005)


def test_FormulaGenerator_generate_formulas_invalid_tolerance():
    config = FormulaGeneratorConfiguration.from_chnops(500)
    formula_generator = fg.FormulaGenerator(config)
    with pytest.raises(ValueError):
        formula_generator.generate_formulas(100, -0.005)


@pytest.mark.parametrize("mass", [500, 1000, 1500, 2000])
def test_FormulaGenerator_from_hmdb(mass):
    FormulaGeneratorConfiguration.from_chnops(mass)


def test_FormulaGenerator_from_hmdb_invalid_mass():
    with pytest.raises(ValueError):
        FormulaGeneratorConfiguration.from_chnops(2450)



import numpy as np
import pytest

from tidyms2.chem import EnvelopeScorer, EnvelopeValidator, Formula
from tidyms2.chem.config import EnvelopeScorerConfiguration, EnvelopeValidatorConfiguration

formula_str_list = ["C11H12N2O2", "C6H12O6", "C27H46O", "CO2", "HCOOH"]


class TestEnvelopeValidator:
    max_length = 5

    @pytest.fixture(scope="class")
    def validator(self):
        config = EnvelopeValidatorConfiguration.from_chnops(500, max_length=self.max_length)
        return EnvelopeValidator(config)

    @pytest.mark.parametrize("f_str", formula_str_list)
    def test_validate_valid_envelope_ok(self, f_str: str, validator: EnvelopeValidator):
        f = Formula(f_str)
        env = f.get_isotopic_envelope(self.max_length)
        validated_length = validator.validate(env.mz, env.p)
        assert validated_length == self.max_length

    def test_validate_invalid_envelope_return_zero_validated_length(self, validator):
        f = Formula("C2H8B")
        env = f.get_isotopic_envelope(self.max_length)
        validated_length = validator.validate(env.mz, env.p)
        expected_length = 0
        assert validated_length == expected_length


class TestEnvelopeScorer:
    max_length = 5
    max_M = 1000.0

    @pytest.fixture(scope="class")
    def scorer(self):
        config = EnvelopeScorerConfiguration.from_chnops(500, max_length=5)
        return EnvelopeScorer(config)

    @pytest.mark.parametrize("f_str", formula_str_list)
    def test_EnvelopeScorer(self, f_str, scorer):
        # test that the best scoring candidate has the same molecular formula
        f = Formula(f_str)
        f_envelope = f.get_isotopic_envelope(self.max_length)
        tolerance = 0.005
        scorer.score(f_envelope.mz, f_envelope.p, tolerance)
        coeff, isotopes, _ = scorer.get_top_results(5)
        expected_coeff = [f.composition[x] for x in isotopes]
        assert np.array_equal(expected_coeff, coeff[0])

    @pytest.mark.parametrize("f_str", formula_str_list)
    def test_EnvelopeScorer_length_gt_scorer_max_length_raises_error(self, scorer, f_str):
        # test that the best scoring candidate has the same molecular formula
        f = Formula(f_str)
        f_envelope = f.get_isotopic_envelope(self.max_length + 1)
        tolerance = 0.005

        with pytest.raises(ValueError):
            scorer.score(f_envelope.mz, f_envelope.p, tolerance)

    @pytest.mark.parametrize("f_str", formula_str_list)
    def test_EnvelopeScorer_custom_scorer(self, f_str, scorer):
        def cosine_scorer(mz1, ab1, mz2, ab2, **scorer_params):
            n1 = np.linalg.norm(ab1)
            n2 = np.linalg.norm(ab2)
            norm = np.linalg.norm(ab1 - ab2)
            cosine = norm / (n1 * n2)
            return 1 - cosine

        scorer_config = EnvelopeScorerConfiguration.from_chnops(500)
        scorer = EnvelopeScorer(scorer_config, scorer=cosine_scorer)

        f = Formula(f_str)
        max_length = 5
        f_envelope = f.get_isotopic_envelope(max_length)
        tolerance = 0.005
        scorer.score(f_envelope.mz, f_envelope.p, tolerance)
        coeff, isotopes, _ = scorer.get_top_results(5)
        expected_coeff = [f.composition[x] for x in isotopes]
        assert np.array_equal(expected_coeff, coeff[0])

    @pytest.mark.parametrize("f_str", ["C2H3N", "N2H4", "C3N3H3"])
    def test_scorer_positive_defects_element_only(self, f_str):
        bounds = {"C": (0, 10), "H": (0, 10), "N": (0, 10)}
        config = EnvelopeScorerConfiguration(bounds=bounds, max_length=self.max_length, max_M=self.max_M)
        scorer = EnvelopeScorer(config)

        f = Formula(f_str)
        f_envelope = f.get_isotopic_envelope(self.max_length)
        tolerance = 0.005
        scorer.score(f_envelope.mz, f_envelope.p, tolerance)
        coeff, isotopes, _ = scorer.get_top_results(5)
        expected_coeff = [f.composition[x] for x in isotopes]
        assert np.array_equal(expected_coeff, coeff[0])

    @pytest.mark.parametrize("f_str", ["CS2", "C2OS2", "C3SO"])
    def test_scorer_negative_elements_only(self, f_str):
        bounds = {"C": (0, 10), "O": (0, 10), "S": (0, 10)}
        config = EnvelopeScorerConfiguration(bounds=bounds, max_length=self.max_length, max_M=self.max_M)
        scorer = EnvelopeScorer(config)

        f = Formula(f_str)
        env = f.get_isotopic_envelope(self.max_length)
        tolerance = 0.001
        scorer.score(env.mz, env.p, tolerance)
        coeff, isotopes, _ = scorer.get_top_results(5)
        expected_coeff = [f.composition[x] for x in isotopes]
        assert np.array_equal(expected_coeff, coeff[0])

    @pytest.mark.parametrize("f_str", ["H2O", "H3PO4", "H2SO4"])
    def test_no_carbon_scorer(self, f_str):
        bounds = {"H": (0, 10), "O": (0, 5), "S": (0, 5), "P": (0, 5)}
        config = EnvelopeScorerConfiguration(bounds=bounds, max_length=self.max_length, max_M=self.max_M)
        scorer = EnvelopeScorer(config)

        f = Formula(f_str)
        env = f.get_isotopic_envelope(self.max_length)
        tolerance = 0.005
        scorer.score(env.mz, env.p, tolerance)
        coeff, isotopes, _ = scorer.get_top_results(5)
        expected_coeff = [f.composition[x] for x in isotopes]
        assert np.array_equal(expected_coeff, coeff[0])



"""SQLAlchemy session management utilities."""

from contextlib import contextmanager
from typing import Generator

from sqlalchemy.orm import Session, sessionmaker


@contextmanager
def create_session(factory: sessionmaker[Session]) -> Generator[Session, None, None]:
    """Get a SQLALchemy session and close connection after usage."""
    with factory() as session:
        transaction = session.begin()
        try:
            yield session
            transaction.commit()
        except Exception as e:
            transaction.rollback()
            raise e



"""Cluster-based feature correspondence utilities."""

from collections import Counter
from functools import partial
from typing import Any, Generator, cast

import numpy as np
import pydantic
from joblib import Parallel, delayed
from scipy.optimize import linear_sum_assignment
from sklearn.cluster import DBSCAN
from sklearn.mixture import GaussianMixture

from ..core.models import Annotation, AnnotationPatch, Sample
from ..core.utils.numpy import FloatArray, IntArray1D


class FeatureCorrespondenceParameters(pydantic.BaseModel):
    """Store match_features parameters."""

    groups: list[str] | None = None
    """If provided, use only samples from these groups to create feature groups."""

    min_fraction: float = 0.25
    """The minimum fraction of samples of a given group in a cluster. If `groups`
    is ``None``, the total number of sample is used to compute the minimum fraction.
    """

    max_deviation: float = 3.0
    """The maximum deviation of a feature from a cluster, measured in numbers of
    standard deviations from the cluster.
    """

    n_jobs: int | None = None
    """Number of jobs to run in parallel. ``None`` means 1 unless in a :obj:`joblib.parallel_backend`
    context. ``-1`` means using all processors."""

    max_sample_overlap: float = pydantic.Field(default=0.25, ge=0.0, le=1.0)
    """If `merge_close_features` is set to ``True``, close features will be
    merged if the number of samples where both features where found is higher
    than the number of samples where at least one of them was found. As an example,
    if one feature group has features detected in samples 1, 2, 3, 4 and the
    other has been detected in samples 1 and 2, they were both detected in
    two out of four samples (sample overlap = 0.5). In this case the features
    will not be merged. On the other hand id, if one feature is detected in
    samples 1, 2, 3, 4, and 5, while the other is detected only in sample 1, in
    this case, both feature groups will be merged into a single group.
    second
    """

    silent: bool = True
    """If ``False``, shows a progress bar."""


def match_features(
    descriptors: dict[str, list[float]],
    annotations: list[Annotation],
    sample_list: list[Sample],
    descriptor_tolerance: dict[str, float],
    params: FeatureCorrespondenceParameters,
) -> list[AnnotationPatch]:
    """Match features across samples using DBSCAN and GMM.

    See the :ref:`user guide <ft-correspondence>` for a detailed description of the algorithm.

    :param descriptors: feature descriptors
    :param tolerance: a dictionary that maps descriptors to tolerance value. Descriptors not included
        in this dictionary are not included in the algorithm.
    :param params: the algorithm parameters
    :return: a list annotation patches with feature group data

    """
    # adapt input to algorithm input
    tolerance = np.array(list(descriptor_tolerance.values()))

    # create feature matrix and scale descriptors
    X = np.vstack([descriptors[x] for x in descriptor_tolerance]).T
    n_features, _ = X.shape

    # scale X using minimum tolerance value
    argmin = np.argmin(tolerance).item()
    X = X * tolerance[argmin] / tolerance

    # create sample array
    sample_ids = [x.sample_id for x in annotations]
    sample_encoded, _ = encode(sample_ids)

    # create group array
    sample_id_to_sample = {x.id: x for x in sample_list}
    groups, group_mapping = encode([sample_id_to_sample[x.sample_id].meta.group for x in annotations])

    # count samples per group
    group_sample_pairs = {(x, y) for x, y in zip(groups, sample_encoded)}
    samples_per_group = Counter([x[0] for x in group_sample_pairs])

    if params.groups:
        include_groups = [group_mapping[x] for x in params.groups if x in group_mapping]
    else:
        include_groups = None

    # DBSCAN clustering
    min_samples = _estimate_dbscan_min_sample(samples_per_group, include_groups, params.min_fraction)
    max_size = 100000
    eps = tolerance[argmin]
    cluster = cluster_dbscan(X, eps, min_samples, max_size)

    # estimate the number of species per DBSCAN cluster
    species_per_cluster = _estimate_gmm_n_components(
        sample_encoded, cluster, groups, samples_per_group, include_groups, params.min_fraction
    )

    # split DBSCAN clusters with multiple species
    cluster_iterator = _get_cluster_iterator(X, cluster, sample_encoded, species_per_cluster)

    func = partial(_split_cluster_worker, max_deviation=params.max_deviation)
    func = delayed(func)
    data = Parallel(n_jobs=params.n_jobs)(func(x) for x in cluster_iterator)
    # TODO: Remove score computation.
    refined_cluster, score = _build_label(data, n_features)

    # TODO: remove parameter
    max_sample_overlap = 0.25
    refined_cluster = merge_feature_groups(X, refined_cluster, sample_ids, eps, max_sample_overlap)
    return [AnnotationPatch(id=x.id, field="group", value=y) for x, y in zip(annotations, refined_cluster)]


def encode(lst: list[str]) -> tuple[IntArray1D, dict[str, int]]:
    """Encode a list as an array of integers."""
    unique = set(lst)
    mapping = {v: k for k, v in enumerate(unique)}
    return np.array([mapping[x] for x in lst]), mapping


def _estimate_dbscan_min_sample(
    samples_per_group: dict[int, int], include_groups: list[int] | None, min_fraction: float
) -> int:
    """Compute the DBSCAN model parameter `min_sample`."""
    if include_groups is None:
        min_samples = round(sum(samples_per_group.values()) * min_fraction)
    else:
        min_samples = sum(samples_per_group.values())
        for k, v in samples_per_group.items():
            if k in include_groups:
                tmp = round(v * min_fraction)
                min_samples = min(tmp, min_samples)
    return min_samples


def cluster_dbscan(X: np.ndarray, eps: float, min_samples: int, max_size: int) -> np.ndarray:
    """Cluster rows of X using the DBSCAN algorithm.

    `X` is split into chunks to reduce memory usage. The split is done in a way
    such that the solution obtained is the same as the solution using `X`.

    Auxiliary function to match_features.

    :param X: m/z and rt values for each feature
    :param eps: Used to build epsilon parameter of DBSCAN
    :param min_samples: parameter to pass to DBSCAN
    :param max_size: maximum number of rows in X. If the number of rows is greater than
        this value, the data is processed in chunks to reduce memory usage.

    :return: the assigned cluster by DBSCAN

    """
    n_rows = X.shape[0]

    if n_rows > max_size:
        # sort X based on the values of the columns and find positions to
        # split into smaller chunks of data.
        x1 = X.T[1]
        sorted_index = np.argsort(x1)
        revert_sorted_index = np.argsort(sorted_index)
        X = X[sorted_index]

        # indices to split X based on max_size
        split_index = np.arange(max_size, n_rows, max_size)

        # find split indices candidates
        # it can be shown that if X is split at one of these points, the
        # points in each one of the chunks are not connected with points in
        # another chunk
        dx = np.diff(x1)
        split_candidates = np.where(dx > eps)[0]
        close_index = np.searchsorted(split_candidates, split_index)

        close_index[-1] = min(split_candidates.size - 1, close_index[-1])
        split_index = split_candidates[close_index] + 1
        split_index = np.hstack((0, split_index, n_rows))
    else:
        split_index = np.array([0, n_rows])
        revert_sorted_index = np.arange(X.shape[0])

    # cluster using DBSCAN on each chunk
    dbscan = DBSCAN(eps=eps, min_samples=min_samples, metric="chebyshev")
    n_chunks = split_index.size - 1
    cluster = np.zeros(X.shape[0], dtype=int)
    cluster_counter = 0
    for k in range(n_chunks):
        start = split_index[k]
        end = split_index[k + 1]
        dbscan.fit(X[start:end, :])
        labels = dbscan.labels_
        n_cluster = (np.unique(labels) >= 0).sum()
        labels[labels >= 0] += cluster_counter
        cluster_counter += n_cluster
        cluster[start:end] = labels

    # revert sort on cluster and X
    if n_rows > max_size:
        cluster = cluster[revert_sorted_index]
    return cluster


def _estimate_gmm_n_components(
    samples: np.ndarray[Any, np.dtype[np.integer]],
    clusters: np.ndarray[Any, np.dtype[np.integer]],
    groups: np.ndarray[Any, np.dtype[np.integer]],
    samples_per_group: dict[int, int],
    include_groups: list[int] | None,
    min_fraction: float,
) -> dict[int, int]:
    """Estimate the number of species in a cluster.

    Auxiliary function to match_features.

    :return: a mapping from cluster label to the number of species estimated.

    """
    if include_groups is None:
        include_groups = list(samples_per_group)
    n_clusters: int = np.max(clusters) + 1
    n_groups = len(include_groups)
    species_array = np.zeros((n_groups, n_clusters), dtype=int)
    # estimate the number of species in a cluster according to each group
    for k, cl in enumerate(include_groups):
        n_samples = samples_per_group[cl]
        n_min = round(n_samples * min_fraction)
        c_mask = groups == cl
        c_samples = samples[c_mask]
        c_clusters = clusters[c_mask]
        c_species = _estimate_n_species_one_group(c_samples, c_clusters, n_min, n_clusters)
        species_array[k, :] = c_species
    # keep the estimation with the highest number of species
    species = species_array.max(axis=0)
    n_species_per_cluster = dict(zip(np.arange(n_clusters), species))
    return n_species_per_cluster  # type: ignore


def _estimate_n_species_one_group(
    samples: np.ndarray, clusters: np.ndarray, min_samples: int, n_clusters: int
) -> np.ndarray:
    """Estimates the number of species in a cluster. Assumes only one group.

    Auxiliary function to _estimate_n_species.

    """
    species = np.zeros(n_clusters, dtype=int)
    for cl in range(n_clusters):
        c_mask = clusters == cl
        c_samples = samples[c_mask]
        # count features per sample in a cluster
        s_unique, s_counts = np.unique(c_samples, return_counts=True)
        # count the number of times a sample has k features
        k_unique, k_counts = np.unique(s_counts, return_counts=True)
        k_mask = k_counts >= min_samples
        k_unique = k_unique[k_mask]
        if k_unique.size:
            species[cl] = k_unique.max()
    return species


def _get_cluster_iterator(
    X: np.ndarray,
    cluster: np.ndarray,
    samples: np.ndarray,
    species_per_cluster: dict[int, int],
) -> Generator[tuple[np.ndarray, np.ndarray, int, np.ndarray], None, None]:
    """Yield the rows of X associated with a cluster.

    Auxiliary function to match_features.

    :yield: a tuple consisting of rows of X associated to a cluster, the sample labels associated to X_c
        and the number of species estimated for the cluster and indices of the rows of `X_c` in `X`

    """
    n_cluster = cluster.max() + 1
    for cl in range(n_cluster):
        n_species = species_per_cluster[cl]
        if n_species > 0:
            index = np.where(cluster == cl)[0]
            X_c = X[index, :]
            samples_c = samples[index]
            yield X_c, samples_c, n_species, index


def split_cluster_gmm(
    X_c: np.ndarray, samples_c: np.ndarray, n_species: int, max_deviation: float
) -> tuple[np.ndarray, np.ndarray]:
    """Process each cluster using GMM.

    Auxiliary function to `match_features`.

    :return: a tuple with the labels and indecisiveness
    """
    # fit GMM
    n_rows = X_c.shape[0]
    if n_rows == 1:
        return samples_c, np.ones(n_rows, dtype=float)

    gmm = GaussianMixture(n_components=n_species, covariance_type="diag")
    gmm.fit(X_c)

    # compute the deviation of the features respect to each cluster
    deviation = _get_deviation(
        X_c,
        cast(np.ndarray[Any, np.dtype[np.floating]], gmm.covariances_),
        cast(np.ndarray[Any, np.dtype[np.floating]], gmm.means_),
    )

    # assign each feature in a sample to component in the GMM minimizing the
    # total deviation in the sample.
    label = -np.ones_like(samples_c)  # by default features are set as noise
    unique_samples = np.unique(samples_c)
    # the indecisiveness is a metric that counts the number of samples
    # where more than one feature can be potentially assigned to a species
    # that is, for each species, the number of rows in deviation with values
    # lower than max_deviation are counted as 1 and zero otherwise.
    # This is done for all samples and the indecisiveness is divided by the
    # number of samples.
    indecisiveness = np.zeros(n_species)
    for s in unique_samples:
        sample_mask = samples_c == s
        sample_deviation = deviation[sample_mask, :]
        # indecisiveness
        count = (sample_deviation < max_deviation).sum(axis=0)
        indecisiveness += count > 1

        # Find the best option for each feature
        best_row, best_col = linear_sum_assignment(sample_deviation)

        # features with deviation greater than max_deviation are set to noise
        valid_ft_mask = sample_deviation[best_row, best_col] <= max_deviation
        best_col = best_col[valid_ft_mask]
        best_row = best_row[valid_ft_mask]
        sample_label = -np.ones(sample_deviation.shape[0], dtype=int)
        sample_label[best_row] = best_col
        label[sample_mask] = sample_label
    indecisiveness /= unique_samples.size
    return label, indecisiveness


def _get_deviation(X: np.ndarray, covariances_: np.ndarray, means_: np.ndarray) -> np.ndarray:
    """Compute the deviation of features.

    Auxiliary function to _process_cluster.

    """
    n_species = covariances_.shape[0]
    n_ft = X.shape[0]
    deviation = np.zeros((n_ft, n_species))
    for k, (m, s) in enumerate(zip(means_, covariances_)):
        # the deviation is the absolute value of X after standardization
        Xs = np.abs((X - m) / np.sqrt(s))
        deviation[:, k] = Xs.max(axis=1)
    return deviation


def _split_cluster_worker(args, max_deviation):
    """Worker used to parallelize feature clustering."""
    Xc, samples_c, n_ft, index = args
    label_c, score = split_cluster_gmm(Xc, samples_c, n_ft, max_deviation)
    return label_c, score, index, n_ft


def _build_label(data, size):
    """Merge the data obtained from each cluster.

    Auxiliary function to match_features.

    """
    label = -1 * np.ones(size, dtype=int)
    cluster_count = 0
    score_list = list()
    for c_label, c_score, c_index, c_n_ft in data:
        c_label[c_label > -1] += cluster_count
        label[c_index] = c_label
        cluster_count += c_n_ft
        score_list.append(c_score)
    score_list = np.hstack(score_list)
    return label, score_list


def _get_progress_bar_total(ft_per_cluster: dict[int, int]) -> int:
    total = 0
    for k, v in ft_per_cluster.items():
        if (k > -1) and (v > 0):
            total += 1
    return total


def merge_feature_groups(
    X: FloatArray, groups: IntArray1D, samples: list[str], tol: float, max_sample_overlap: float
) -> IntArray1D:
    """Merge close features.

    Feature groups are merged if the two following conditions are meet:

    1.  Each feature group centroid must be closer than the corresponding tolerance.
    2.  The overlap between feature groups must be lower than the `max_sample_overlap` parameter.

    Groups are checked an merged in a pairwise fashion. If a feature group appear
    in more than one pair, then only one is considered. Then after all features
    groups are checked, the algorithm is run until candidate pairs are no longer found.

    """
    unique_groups, centroids = _compute_feature_group_centroids(X, groups)
    group_to_samples = _create_group_to_samples_dict(groups, samples)

    candidates = _find_merge_candidates(centroids, unique_groups, tol)

    if not candidates:
        return groups

    merge_map: dict[int, int] = dict()

    for group1, group2 in candidates:
        if group1 in merge_map or group2 in merge_map:
            continue

        samples1 = group_to_samples[group1]
        samples2 = group_to_samples[group2]

        merger, merged = _merge_if_close(group1, samples1, group2, samples2, max_sample_overlap)

        if merger > -1 and merged > -1:
            merge_map[merged] = merger

    groups = np.array([merge_map[x] if x in merge_map else x for x in groups])
    return merge_feature_groups(X, groups, samples, tol, max_sample_overlap)


def _compute_feature_group_centroids(X: FloatArray, groups: IntArray1D) -> tuple[IntArray1D, FloatArray]:
    """Compute the mean value for each descriptor in a feature group."""
    sorted_index = np.argsort(groups)

    X = X[sorted_index]
    groups = groups[sorted_index]

    # split X into so rows belong to the same feature group
    unique_groups, group_sections = np.unique(groups, return_index=True)

    if unique_groups.size:
        group_sections = group_sections[1:]  # remove first index so it matches split expected format
        centroids = np.vstack([x.mean(axis=0) for x in np.split(X, group_sections)])
    else:
        # TODO: check if this is the best approach for the empty case
        centroids = np.array([], dtype=X.dtype)
    return unique_groups, centroids


def _find_merge_candidates(centroids: FloatArray, groups: IntArray1D, tol: float) -> list[tuple[int, int]]:
    """Group a pair of feature groups if all of their descriptor centroids are closer than the tolerance."""
    sorted_centroids_index = np.lexsort(centroids.T)
    groups = groups[sorted_centroids_index]
    rows_close = np.abs(np.diff(centroids[sorted_centroids_index], axis=0)) < tol
    diff_rows = np.all(rows_close, axis=1)

    candidates = list()
    for i in np.where(diff_rows)[0]:
        candidates.append((groups[i], groups[i + 1]))
    return candidates


def _create_group_to_samples_dict(groups: IntArray1D, samples: list[str]) -> dict[int, set[str]]:
    """Create a dictionary that maps feature groups to samples where a feature was detected."""
    samples_per_group = dict()

    for g, s in zip(groups, samples):
        g_samples = samples_per_group.setdefault(g, set())
        g_samples.add(s)

    return samples_per_group


def _merge_if_close(
    group1: int, samples1: set[str], group2: int, samples2: set[str], max_sample_overlap: float
) -> tuple[int, int]:
    """Flag a pair of feature groups for merging if the sample overlap is lower than maximum specified.

    Returns
    -------
    merger : int
        The feature group label that will be kept after the merge. If no merge is
        is to be made, this value is set to ``-1``.
    merged : int
        The feature group label that will be deleted after the merge. If no merge is
        is to be made, this value is set to ``-1``.

    """
    total = len(samples1.union(samples2))
    n_overlap = len(samples1.intersection(samples2))

    if (n_overlap / total) > max_sample_overlap:
        return (-1, -1)
    return (group1, group2) if len(samples1) >= len(samples2) else (group2, group1)



"""SQLite based storage classes."""

from typing import Generic, Iterable
from uuid import UUID

from sqlalchemy import create_engine, select, update
from sqlalchemy.orm import Session, sessionmaker
from sqlalchemy.sql import text

from ...core import exceptions, models
from ...core.dataflow import AssayProcessStatus, SampleProcessStatus
from ...core.models import (
    Annotation,
    AnnotationPatch,
    DescriptorPatch,
    FeatureGroup,
    FeatureType,
    FillValue,
    RoiType,
    Sample,
)
from ...core.storage import SampleStorage
from ..memory import OnMemorySampleStorage
from ..table import FeatureTable
from .models import (
    AnnotationModel,
    AnnotationPatchModel,
    AssaySnapshotStatusModel,
    Base,
    DescriptorModel,
    DescriptorPatchModel,
    FeatureGroupModel,
    FeatureModel,
    FillValueModel,
    RoiModel,
    SampleModel,
    SampleSnapshotModel,
)
from .session import create_session

LATEST_SNAPSHOT = "head"


class SQLiteAssayStorage(Generic[FeatureType, RoiType]):
    """Assay storage class for that persists data using a SQLite backend.

    :param id: an identifier for the storage
    :param host: the DB host string. If not provided an in-memory database is used
    :param roi_type: the ROI class stored in the DB.
    :param feature_type: the feature class stored in the DB.

    """

    def __init__(self, id: str, host: str | None, roi_type: type[RoiType], feature_type: type[FeatureType]):
        self.id = id
        self.roi_type = roi_type
        self.feature_type = feature_type
        self._table: FeatureTable | None = None
        self._current_snapshot_id = LATEST_SNAPSHOT

        if host is None:
            host = "sqlite:///:memory:"
        else:
            host = f"sqlite:///{host}"

        self.engine = create_engine(host)
        self.session_factory = sessionmaker(bind=self.engine)

        Base.metadata.create_all(self.engine)

        # create a snapshot if it does not exist
        stmt = select(AssaySnapshotStatusModel).where(AssaySnapshotStatusModel.id == self._current_snapshot_id)
        with create_session(self.session_factory) as session:
            status = session.execute(stmt).scalar()
            if status is None:
                initial_state = AssayProcessStatus()
                self._add_assay_snapshot_meta(session, LATEST_SNAPSHOT, initial_state)

    def add_feature_groups(self, *groups: FeatureGroup) -> None:
        """Add feature groups data to the assay storage."""
        with create_session(self.session_factory) as session:
            session.add_all(FeatureGroupModel.from_feature_group(x) for x in groups)

    def add_fill_values(self, *fill_values: FillValue) -> None:
        """Add values to fill missing entries in the data matrix."""
        table = self._fetch_feature_table()
        d = dict()
        for fill in fill_values:
            sample_d = d.setdefault(fill.sample_id, dict())
            sample_d[fill.feature_group] = fill.value
        table.add_fill_values(d)

        snapshot_id = self.get_snapshot_id()

        with create_session(self.session_factory) as session:
            session.add_all(FillValueModel.from_fill_value(x, snapshot_id) for x in fill_values)

    def add_sample_data(self, data: SampleStorage[RoiType, FeatureType]) -> None:
        """Add sample data to DB."""
        with create_session(self.session_factory) as session:
            sample = data.get_sample()
            if self.has_sample(sample.id):
                raise exceptions.RepeatedIdError(f"Sample with id {sample.id} already stored in the assay.")
            self._add_sample(session, sample)
            self._add_rois(session, data)
            self._add_features(session, data)
            self._add_sample_snapshot_meta(session, data)

    def create_snapshot(self, snapshot_id: str) -> None:
        """Create a new assay data snapshot."""
        self._check_current_snapshot_is_latest_snapshot()
        with create_session(self.session_factory) as session:
            current = self._fetch_current_status(session)
            new_snap = AssaySnapshotStatusModel(id=LATEST_SNAPSHOT, status=self.get_process_status().model_dump_json())
            current.id = snapshot_id
            session.add(new_snap)

            annotation_update = (
                update(AnnotationPatchModel)
                .where(AnnotationPatchModel.snapshot_id == LATEST_SNAPSHOT)
                .values(snapshot_id=snapshot_id)
            )
            session.execute(annotation_update)

            descriptor_update = (
                update(DescriptorPatchModel)
                .where(DescriptorPatchModel.snapshot_id == LATEST_SNAPSHOT)
                .values(snapshot_id=snapshot_id)
            )
            session.execute(descriptor_update)

            fill_value_update = (
                update(FillValueModel)
                .where(FillValueModel.snapshot_id == LATEST_SNAPSHOT)
                .values(snapshot_id=snapshot_id)
            )
            session.execute(fill_value_update)

    def fetch_annotations(self, sample_id: str | None = None) -> list[Annotation]:
        """Fetch the feature annotations."""
        table = self._fetch_feature_table()
        return table.fetch_annotations(sample_id=sample_id, copy=True)

    def fetch_descriptors(self, sample_id: str | None = None, descriptors: Iterable[str] | None = None) -> dict:
        """Fetch the feature descriptors."""
        table = self._fetch_feature_table()
        all_descriptors = self.feature_type.descriptor_names()
        if descriptors is None:
            descriptors = list(all_descriptors)
        for x in descriptors:
            if x not in all_descriptors:
                raise exceptions.InvalidFeatureDescriptor(x)
        return table.fetch_descriptors(descriptors=descriptors, sample_id=sample_id, copy=True)

    def fetch_features_by_group(self, group: int) -> list[FeatureType]:
        """Fetch features using the feature group id."""
        stmt = (
            select(FeatureModel, AnnotationModel)
            .join(AnnotationModel, FeatureModel.id == AnnotationModel.id)
            .where(AnnotationModel.group == group)
            .where(FeatureModel.snapshot_id == LATEST_SNAPSHOT)
        )
        with create_session(self.session_factory) as session:
            results = [row._tuple() for row in session.execute(stmt).all()]
            features = self._fetch_features_from_model(session, results)

        return features

    def fetch_features_by_id(self, *feature_ids: UUID) -> list[FeatureType]:
        """Fetch features using their ids."""
        stmt = (
            select(FeatureModel, AnnotationModel)
            .join(AnnotationModel, FeatureModel.id == AnnotationModel.id)
            .where(FeatureModel.id.in_(feature_ids))
            .where(FeatureModel.snapshot_id == LATEST_SNAPSHOT)
        )
        with create_session(self.session_factory) as session:
            results = [row._tuple() for row in session.execute(stmt).all()]
            features = self._fetch_features_from_model(session, results)

        return features

    def fetch_features_by_sample(self, sample_id: str) -> list[FeatureType]:
        """Fetch features using the sample id."""
        stmt = (
            select(FeatureModel, AnnotationModel)
            .join(AnnotationModel, FeatureModel.id == AnnotationModel.id)
            .where(AnnotationModel.sample_id == sample_id)
            .where(FeatureModel.snapshot_id == LATEST_SNAPSHOT)
        )
        with create_session(self.session_factory) as session:
            results = [row._tuple() for row in session.execute(stmt).all()]
            features = self._fetch_features_from_model(session, results)

        return features

    def fetch_feature_groups(self) -> list[FeatureGroup]:
        """Fetch feature groups stored in the assay."""
        with create_session(self.session_factory) as session:
            stmt = select(FeatureGroupModel)
            feature_groups = [x.to_feature_group() for x in session.execute(stmt).scalars()]
        return feature_groups

    def fetch_fill_values(self) -> dict[str, dict[int, float]]:
        """Fetch fill values for data matrix."""
        table = self._fetch_feature_table()
        return table.fetch_fill_values(copy=True)

    def fetch_rois_by_id(self, *roi_ids: UUID) -> list[RoiType]:
        """Fetch ROIs using their ids.

        :param roi_ids: a list of ROI ids to fetch

        """
        with create_session(self.session_factory) as session:
            rois = self._fetch_rois_by_id(session, LATEST_SNAPSHOT, *roi_ids)
        return rois

    def fetch_rois_by_sample(self, sample_id: str) -> list[RoiType]:
        """Fetch all ROIs from a sample."""
        with create_session(self.session_factory) as session:
            sample = self._fetch_sample(session, sample_id)
            stmt = (
                select(RoiModel)
                .where((RoiModel.sample_id == sample_id))
                .where(RoiModel.snapshot_id == LATEST_SNAPSHOT)
            )
            models = session.execute(stmt).scalars()
            rois = [self.roi_type.from_str(x.data, sample) for x in models]
        return rois

    def get_feature_type(self) -> type[FeatureType]:
        """Retrieve the Feature class used."""
        return self.feature_type

    def get_roi_type(self) -> type[RoiType]:
        """Retrieve the ROI class used."""
        return self.roi_type

    def get_n_rois(self) -> int:
        """Retrieve the number of ROIs in the assay."""
        with create_session(self.session_factory) as session:
            stmt = text("SELECT COUNT(1) FROM rois;")
            n_rois = session.execute(stmt).scalar_one_or_none()
        n_rois = 0 if n_rois is None else n_rois
        return n_rois

    def get_n_features(self) -> int:
        """Retrieve the number of features in the assay."""
        with create_session(self.session_factory) as session:
            stmt = text("SELECT COUNT(1) FROM features;")
            n_fts = session.execute(stmt).scalar_one_or_none()
        n_fts = 0 if n_fts is None else n_fts
        return n_fts

    def get_n_samples(self) -> int:
        """Retrieve the number of samples in the assay."""
        with create_session(self.session_factory) as session:
            stmt = text("SELECT COUNT(1) FROM samples;")
            n_samples = session.execute(stmt).scalar_one_or_none()
        n_samples = 0 if n_samples is None else n_samples
        return n_samples

    def get_process_status(self) -> AssayProcessStatus:
        """Get the current process status."""
        with create_session(self.session_factory) as session:
            status = self._fetch_current_status(session).to_status()
        return status

    def get_snapshot_id(self) -> str:
        """Retrieve the current snapshot id."""
        return self._current_snapshot_id

    def fetch_sample(self, sample_id: str) -> Sample:
        """Retrieve a sample from the assay.

        :param sample_id: the id of the sample to retrieve
        :raises SampleNotFound: if the provided id is not found in the DB

        """
        with create_session(self.session_factory) as session:
            sample = self._fetch_sample(session, sample_id)
        return sample

    def fetch_sample_data(self, sample_id: str) -> OnMemorySampleStorage[RoiType, FeatureType]:
        """Fetch Samples from the assay using their ids."""
        with create_session(self.session_factory) as session:
            sample = self._fetch_sample(session, sample_id)
            all_rois = self._fetch_all_rois(session, sample)
            all_features = self._fetch_all_features(session, sample, all_rois)
            snapshot_ids = self._fetch_sample_snapshot_ids(session, sample.id)
            states = self._fetch_sample_states(session, sample.id)
            data = OnMemorySampleStorage.from_dict(
                sample, all_rois, all_features, snapshot_ids, states, self.roi_type, self.feature_type
            )

        return data

    def has_feature(self, feature_id: UUID) -> bool:
        """Check if a feature with the provided id is stored in the DB.

        :param feature_id: the id of the ROI to check

        """
        with create_session(self.session_factory) as session:
            stmt = select(FeatureModel).where(FeatureModel.id == feature_id)
            check = session.execute(stmt).scalars().first() is not None
        return check

    def has_feature_group(self, feature_group: int) -> bool:
        """Check if a group with the provided id is in the assay."""
        raise NotImplementedError

    def has_roi(self, roi_id: UUID) -> bool:
        """Check if a ROI with the provided id is stored in the DB.

        :param roi_id: the id of the ROI to check

        """
        with create_session(self.session_factory) as session:
            stmt = select(RoiModel).where(RoiModel.id == roi_id)
            check = session.execute(stmt).scalars().first() is not None
        return check

    def has_sample(self, sample_id: str) -> bool:
        """Check if a sample with the provided id is stored in the DB.

        :param sample_id: the id of the sample to check

        """
        with create_session(self.session_factory) as session:
            stmt = select(SampleModel).where(SampleModel.id == sample_id)
            check = session.execute(stmt).scalars().first() is not None
        return check

    def list_samples(self) -> list[Sample]:
        """List samples in the assay."""
        with create_session(self.session_factory) as session:
            samples = self._fetch_all_samples(session)
        return samples

    def list_feature_groups(self) -> list[int]:
        """List all group ids in the assay."""
        raise NotImplementedError

    def list_snapshots(self) -> list[str]:
        """Retrieve the list of all snapshots."""
        with create_session(self.session_factory) as session:
            snaps = self._list_snapshots(session)
        return snaps

    def _list_snapshots(self, session: Session) -> list[str]:
        """Retrieve the list of all snapshots."""
        stmt = select(AssaySnapshotStatusModel.id).order_by(AssaySnapshotStatusModel.order)
        return [x for x in session.execute(stmt).scalars()]

    def patch_annotations(self, *patches: models.AnnotationPatch) -> None:
        """Update feature annotation values."""
        self._check_current_snapshot_is_latest_snapshot()
        table = self._fetch_feature_table()
        table.patch_annotation(*patches)
        with create_session(self.session_factory) as session:
            self._add_annotation_patches(session, *patches)

    def patch_descriptors(self, *patches: models.DescriptorPatch) -> None:
        """Update feature descriptors values."""
        self._check_current_snapshot_is_latest_snapshot()
        table = self._fetch_feature_table()
        table.patch_descriptors(*patches)
        with create_session(self.session_factory) as session:
            self._add_descriptor_patches(session, *patches)

    def set_process_status(self, status: AssayProcessStatus) -> None:
        """Set the new process status."""
        self._check_current_snapshot_is_latest_snapshot()
        with create_session(self.session_factory) as session:
            stmt = select(AssaySnapshotStatusModel).where(AssaySnapshotStatusModel.id == self._current_snapshot_id)
            current = session.execute(stmt).scalar_one()
            current.status = status.model_dump_json()

    def set_snapshot(self, snapshot_id: str | None = None, reset: bool = False) -> None:
        """Set assay storage data to specified snapshot.

        If ``None``, fetch data from the latest snapshot.

        """
        if snapshot_id is None:
            snapshot_id = LATEST_SNAPSHOT

        if snapshot_id == self._current_snapshot_id and not reset:
            return

        stmt = select(AssaySnapshotStatusModel).where(AssaySnapshotStatusModel.id == snapshot_id)
        with create_session(self.session_factory) as session:
            model = session.execute(stmt).scalar()
            if model is None:
                raise exceptions.SnapshotNotFound(snapshot_id)
            self._current_snapshot_id = snapshot_id
        self._table = None  # remove cached feature descriptor and annotation data

    def _add_sample(self, session: Session, sample: Sample) -> None:
        """Add sample to DB."""
        model = SampleModel.from_sample(sample)
        session.add(model)

    def _add_features(self, session: Session, data: SampleStorage[RoiType, FeatureType]) -> None:
        feature_models = list()
        ann_models = list()
        descriptor_models = list()
        for snapshot_id in data.list_snapshots():
            data.set_snapshot(snapshot_id)
            snapshot_features = data.list_features()
            feature_models.extend([FeatureModel.from_feature(x, snapshot_id) for x in snapshot_features])
            ann_models.extend([AnnotationModel.from_feature(x, snapshot_id) for x in snapshot_features])
            descriptor_models.extend([DescriptorModel.from_feature(x, snapshot_id) for x in snapshot_features])
        data.set_snapshot()
        session.add_all(descriptor_models)
        session.add_all(feature_models)
        session.add_all(ann_models)

    def _add_rois(self, session: Session, data: SampleStorage[RoiType, FeatureType]) -> None:
        roi_models = list()
        for snapshot_id in data.list_snapshots():
            data.set_snapshot(snapshot_id)
            roi_models.extend([RoiModel.from_roi(x, snapshot_id) for x in data.list_rois()])
        data.set_snapshot()
        session.add_all(roi_models)

    def _add_sample_snapshot_meta(self, session: Session, data: SampleStorage[RoiType, FeatureType]) -> None:
        id_ = data.get_sample().id
        snapshots = list()
        current_snapshot_id = data.get_snapshot_id()
        for k, snapshot_id in enumerate(data.list_snapshots()):
            data.set_snapshot(snapshot_id)
            state = data.get_status()
            snap = SampleSnapshotModel(sample_id=id_, order=k, snapshot_id=snapshot_id, status=state.model_dump_json())
            snapshots.append(snap)
        data.set_snapshot(current_snapshot_id)
        session.add_all(snapshots)

    def _add_annotation_patches(self, session: Session, *patches: AnnotationPatch) -> None:
        session.add_all(AnnotationPatchModel.from_patch(x, snapshot=LATEST_SNAPSHOT) for x in patches)

    def _add_descriptor_patches(self, session: Session, *patches: DescriptorPatch) -> None:
        session.add_all(DescriptorPatchModel.from_patch(x, snapshot=LATEST_SNAPSHOT) for x in patches)

    def _check_current_snapshot_is_latest_snapshot(self):
        id_ = self._current_snapshot_id
        if id_ != LATEST_SNAPSHOT:
            msg = f"Only latest snapshot can be modified. Cannot change the state of snapshot {id_}"
            raise exceptions.SnapshotError(msg)

    def _fetch_sample_snapshot_ids(self, session: Session, sample_id: str) -> list[str]:
        stmt = (
            select(SampleSnapshotModel.snapshot_id)
            .where(SampleSnapshotModel.sample_id == sample_id)
            .order_by(SampleSnapshotModel.order)
        )
        return list(session.execute(stmt).scalars().all())

    def _fetch_sample_states(self, session: Session, sample_id: str) -> dict[str, SampleProcessStatus]:
        stmt = select(SampleSnapshotModel).where(SampleSnapshotModel.sample_id == sample_id)
        results = session.execute(stmt).scalars().all()
        return {x.snapshot_id: SampleProcessStatus.model_validate_json(x.status) for x in results}

    def _fetch_features_from_model(self, session: Session, results: list[tuple[FeatureModel, AnnotationModel]]):
        roi_ids = {x[0].roi_id for x in results}
        id_to_roi = {x.id: x for x in self._fetch_rois_by_id(session, LATEST_SNAPSHOT, *roi_ids)}
        features = list()
        for ft_model, ann_model in results:
            roi = id_to_roi[ann_model.roi_id]
            ft = self.feature_type.from_str(ft_model.data, roi, ann_model.to_dict())  # type: ignore
            features.append(ft)
        return features

    def _fetch_all_samples(self, session: Session) -> list[Sample]:
        stmt = select(SampleModel)
        return [x.to_pydantic_model() for x in session.execute(stmt).scalars()]

    def _fetch_sample(self, session: Session, sample_id: str) -> Sample:
        stmt = select(SampleModel).where(SampleModel.id == sample_id)
        model = session.execute(stmt).scalars().first()
        if model is None:
            raise exceptions.SampleNotFound(sample_id)
        return model.to_pydantic_model()

    def _fetch_rois_by_id(self, session: Session, snapshot_id: str, *roi_ids: UUID) -> list[RoiType]:
        stmt = select(RoiModel).where(RoiModel.id.in_(roi_ids)).where(RoiModel.snapshot_id == snapshot_id)
        sample_id_to_sample = {x.id: x for x in self._fetch_all_samples(session)}
        roi_models = session.execute(stmt).scalars()
        rois = [self.roi_type.from_str(x.data, sample_id_to_sample[x.sample_id]) for x in roi_models]
        return rois

    def _fetch_all_rois(self, session: Session, sample: Sample) -> dict[str, list[RoiType]]:
        """Fetch ROIs from all snapshots."""
        stmt = select(RoiModel).where(RoiModel.sample_id == sample.id)
        snapshot_to_rois: dict[str, list[RoiType]] = dict()
        for roi_model in session.execute(stmt).scalars():
            rois = snapshot_to_rois.setdefault(roi_model.snapshot_id, list())
            rois.append(self.roi_type.from_str(roi_model.data, sample))
        return snapshot_to_rois

    def _fetch_all_features(self, session: Session, sample: Sample, snapshot_to_roi: dict[str, list[RoiType]]):
        """Fetch features from all snapshots."""
        stmt = (
            select(FeatureModel, AnnotationModel)
            .join(AnnotationModel, FeatureModel.id == AnnotationModel.id)
            .where(AnnotationModel.sample_id == sample.id)
        )
        snapshot_to_features: dict[str, list[FeatureType]] = dict()

        roi_id_to_roi = dict()
        for roi_list in snapshot_to_roi.values():
            for roi in roi_list:
                roi_id_to_roi[roi.id] = roi

        for row in session.execute(stmt).all():
            ft_model, ann_model = row._tuple()
            features = snapshot_to_features.setdefault(ft_model.snapshot_id, list())

            annotation = Annotation(**ann_model.to_dict())
            roi = roi_id_to_roi[annotation.roi_id]
            features.append(self.feature_type.from_str(ft_model.data, roi, annotation))

        return snapshot_to_features

    def _fetch_feature_table(self) -> FeatureTable:
        if self._table is not None:
            return self._table

        with create_session(self.session_factory) as session:
            descriptors = self._fetch_descriptors(session, LATEST_SNAPSHOT)
            annotations = self._fetch_annotations(session, LATEST_SNAPSHOT)
            ann_patches = self._fetch_annotations_patches(session)
            descriptor_patches = self._fetch_descriptor_patches(session)
            fill_values = self._fetch_fill_values(session)
        self._table = FeatureTable()
        self._table.add_descriptors(descriptors, annotations)
        self._table.add_fill_values(fill_values)
        for patches in ann_patches:
            self._table.patch_annotation(*patches)
        for patches in descriptor_patches:
            self._table.patch_descriptors(*patches)
        return self._table

    def _fetch_descriptors(self, session: Session, snapshot_id: str) -> list[dict[str, float]]:
        stmt = select(DescriptorModel).where(DescriptorModel.snapshot_id == snapshot_id)
        return [x.create_descriptor_dict() for x in session.execute(stmt).scalars()]

    def _fetch_annotations(self, session: Session, snapshot_id: str) -> list[Annotation]:
        stmt = select(AnnotationModel).where(AnnotationModel.snapshot_id == snapshot_id)
        return [x.to_annotation() for x in session.execute(stmt).scalars()]

    def _fetch_annotations_patches(self, session: Session) -> list[list[AnnotationPatch]]:
        current_snapshot = self.get_snapshot_id()
        all_snapshots = self._list_snapshots(session)
        current_index = all_snapshots.index(current_snapshot)
        fetch_snapshots = all_snapshots[: current_index + 1]

        stmt = select(AnnotationPatchModel).where(AnnotationPatchModel.snapshot_id.in_(fetch_snapshots))

        patches = dict()
        for p in session.execute(stmt).scalars():
            snapshot_patches = patches.setdefault(p.snapshot_id, list())
            snapshot_patches.append(AnnotationPatch(id=p.feature_id, field=p.field, value=p.value))

        return [patches.get(x, list()) for x in fetch_snapshots]

    def _fetch_descriptor_patches(self, session: Session) -> list[list[DescriptorPatch]]:
        current_snapshot = self.get_snapshot_id()
        all_snapshots = self._list_snapshots(session)
        current_index = all_snapshots.index(current_snapshot)
        fetch_snapshots = all_snapshots[: current_index + 1]

        stmt = select(DescriptorPatchModel).where(DescriptorPatchModel.snapshot_id.in_(fetch_snapshots))

        patches = dict()
        for p in session.execute(stmt).scalars():
            snapshot_patches = patches.setdefault(p.snapshot_id, list())
            snapshot_patches.append(DescriptorPatch(id=p.feature_id, descriptor=p.descriptor, value=p.value))

        return [patches.get(x, list()) for x in fetch_snapshots]

    def _fetch_fill_values(self, session: Session) -> dict[str, dict[int, float]]:
        current_snapshot = self.get_snapshot_id()
        all_snapshots = self._list_snapshots(session)
        current_index = all_snapshots.index(current_snapshot)
        fetch_snapshots = all_snapshots[: current_index + 1]

        stmt = select(FillValueModel).where(FillValueModel.snapshot_id.in_(fetch_snapshots))

        fill_values = dict()
        for fill in session.execute(stmt).scalars():
            sample_fills = fill_values.setdefault(fill.sample_id, dict())
            sample_fills.setdefault(fill.feature_group, 0.0)
            sample_fills[fill.feature_group] += fill.value
        return fill_values

    def _fetch_current_status(self, session: Session) -> AssaySnapshotStatusModel:
        stmt = select(AssaySnapshotStatusModel).where(AssaySnapshotStatusModel.id == self._current_snapshot_id)
        return session.execute(stmt).scalar_one()

    def _add_assay_snapshot_meta(self, session: Session, snapshot_id: str, state: AssayProcessStatus) -> None:
        model = AssaySnapshotStatusModel(id=snapshot_id, status=state.model_dump_json())
        session.add(model)



"""TidyMS core exceptions."""


class UnprocessedSampleError(ValueError):
    """Exception raised when an assay processing pipeline is applied on an assay with queued sample data."""


class FeatureGroupNotFound(ValueError):
    """Exception raised when a feature group is not found in an assay storage."""


class FeatureNotFound(ValueError):
    """Exception raised when a Feature is not found in a sample storage or assay storage."""


class InvalidFeatureDescriptor(ValueError):
    """Exception raised when a non existing feature descriptor name is requested."""


class PipelineConfigurationError(ValueError):
    """Exception raised when an invalid configuration is set in a pipeline."""


class RepeatedIdError(ValueError):
    """Exception raised when trying to add a resource with an existing id."""


class RoiNotFound(ValueError):
    """Exception raised when a Roi is not found in a sample storage or assay storage."""


class ProcessStatusError(ValueError):
    """Exception raised when an action cannot be performed on sample data due to incorrect processing status."""


class RegistryError(ValueError):
    """Exception raised when an entry is not found in a registry."""


class SampleNotFound(ValueError):
    """Exception raised when a sample is not found in assay/matrix storage."""


class SnapshotNotFound(ValueError):
    """Exception raised when a snapshot is not found in the sample or assay storage."""


class SnapshotError(ValueError):
    """Exception raised from snapshot data related errors."""


class DatasetNotFound(ValueError):
    """Exception raised when a dataset is not found."""


class InvalidSeparationMode(ValueError):
    """Exception raised when an invalid separation mode is passed."""


class SampleProcessorError(ValueError):
    """Exception raised when a sample processor fails to process a sample."""


class RepeatedSampleOrder(ValueError):
    """Exception raised when multiple samples have the same order."""


class EmptyDataMatrix(ValueError):
    """Exception raised when an empty data matrix is created or a transformation results in an empty matrix."""


class MatrixQueryError(ValueError):
    """Exception raised when an invalid sample query is created."""


class SampleMetadataNotFound(ValueError):
    """Exception raises when a sample does not contain required metadata."""



"""Tools for working with Isotopes and Elements."""

from __future__ import annotations

from functools import cached_property
from math import isclose
from typing import Any, Sequence

import pydantic

from .exceptions import IsotopeError

EM = 0.00054858  # electron mass


class Isotope(pydantic.BaseModel, frozen=True):
    """Store Isotope mass and abundance information."""

    z: pydantic.PositiveInt
    """The atomic number"""

    a: pydantic.PositiveInt
    """The atomic mass number"""

    m: pydantic.PositiveFloat = pydantic.Field(repr=False)
    """The exact mass"""

    symbol: str
    """The element symbol"""

    p: float = pydantic.Field(ge=0.0, le=1.0, repr=False)
    """The isotope abundance"""

    @pydantic.computed_field(repr=False)
    @cached_property
    def n(self) -> int:
        """The number of neutrons."""
        return self.a - self.z

    @pydantic.computed_field(repr=False)
    @cached_property
    def d(self) -> float:
        """The mass defect."""
        return self.m - self.a

    def to_str(self) -> str:
        """Create a string representation of the isotope."""
        return f"{self.a}{self.symbol}"


class Element(pydantic.BaseModel, frozen=True):
    """Store element information and its isotopes."""

    name: str = pydantic.Field(repr=False)
    """The element name"""

    symbol: str
    """The element symbol"""

    isotopes: Sequence[Isotope] = pydantic.Field(repr=False)
    """Maps mass numbers to isotopes"""

    z: pydantic.PositiveInt
    """The atomic number"""

    @pydantic.field_validator("isotopes")
    @classmethod
    def _validate_isotope_abundances(cls, isotopes: Sequence[Isotope]) -> Sequence[Isotope]:
        total = sum(x.p for x in isotopes)
        if not isclose(total, 1.0):
            z = isotopes[0].z
            raise IsotopeError(
                f"The total isotope abundance of an element must be equal to 1.0. Got {total} for z={z}"
            )
        return isotopes

    @pydantic.model_validator(mode="before")
    @classmethod
    def _validate_isotopes_z(cls, data: Any) -> Any:
        z = data.get("z")
        assert isinstance(z, int)
        if not all(z == x.z for x in data["isotopes"]):
            raise IsotopeError("All isotopes in an element must have the same atomic number.")
        return data

    @pydantic.computed_field(repr=False)
    @cached_property
    def mmi(self) -> Isotope:
        """Return the isotope with the lowest atomic mass."""
        return min(self.isotopes, key=lambda x: x.a)

    @pydantic.computed_field(repr=False)
    @cached_property
    def monoisotope(self) -> Isotope:
        """Return the most abundant isotope."""
        return max(self.isotopes, key=lambda x: x.p)



"""TidyMS core data models.

All models defined here define how data is shared in data pipelines.

Refer to the :ref:`architecture overview <overview>` for an introduction to the TidyMS data model.

For considerations on customizing these models, refer to the :ref:`extending-guide` guide.

"""

from __future__ import annotations

import json
from abc import ABC, abstractmethod
from functools import cached_property, lru_cache
from pathlib import Path
from typing import Generic, Self, TypeVar
from uuid import UUID

import numpy
import pydantic

from . import exceptions
from .enums import MSDataMode
from .utils.common import create_id
from .utils.numpy import FloatArray1D, IntArray1D


class TidyMSBaseModel(pydantic.BaseModel):
    """Base model that all other library models inherit from."""

    id: UUID = pydantic.Field(default_factory=create_id)
    """A unique id for the model."""

    model_config = pydantic.ConfigDict(validate_assignment=True, arbitrary_types_allowed=True)


class SampleMetadata(pydantic.BaseModel):
    """Sample metadata container."""

    model_config = pydantic.ConfigDict(validate_assignment=True, extra="allow")

    type: str = ""
    """the :term:`sample type`"""

    group: str = ""
    """the :term:`sample group`"""

    order: pydantic.NonNegativeInt = 0
    """the sample measurement order in an assay"""

    batch: pydantic.NonNegativeInt = pydantic.Field(default=0, repr=False)
    """the sample analytical batch number in an assay."""


class Sample(pydantic.BaseModel):
    """Store information required to load data from a raw data file."""

    model_config = pydantic.ConfigDict(validate_assignment=True, frozen=True, extra="forbid")

    path: Path
    """Path to a raw data file"""

    id: str
    """A unique sample identifier"""

    ms_level: pydantic.PositiveInt = pydantic.Field(default=1, repr=False)
    """the sample MS level"""

    reader: str | None = pydantic.Field(default=None, repr=False)
    """The name of a registered data reader to read sample data. If ``None``, the
    optimal reader is inferred from the file extension.
    """

    ms_data_mode: MSDataMode = pydantic.Field(default=MSDataMode.CENTROID, repr=False)
    """the mode in which the sample data is stored."""

    start_time: pydantic.NonNegativeFloat = pydantic.Field(default=0.0, repr=False)
    """Minimum acquisition time of MS scans to include. If ``None``, start from the first scan"""

    end_time: pydantic.NonNegativeFloat | None = pydantic.Field(default=None, repr=False)
    """Maximum acquisition time of MS scans to include. If ``None``, end at the last scan"""

    meta: SampleMetadata = pydantic.Field(default=SampleMetadata(), repr=False)
    """Sample metadata."""

    @pydantic.field_validator("path", mode="before")
    def _normalize_path(cls, value: Path) -> Path:
        if isinstance(value, str):
            value = Path(value)
        return value

    @pydantic.field_serializer("path")
    def serialize_path(self, path: Path, _info) -> str:
        """Serialize path into a string."""
        return str(path)

    @staticmethod
    def validate_samples(*samples: Sample) -> None:
        n_samples = len(samples)
        if len({x.id for x in samples}) < n_samples:
            msg = "Samples must have a unique id."
            raise exceptions.RepeatedIdError(msg)

        if len({x.meta.order for x in samples}) < n_samples:
            msg = "Samples must have a unique order."
            raise exceptions.RepeatedSampleOrder(msg)


class Roi(TidyMSBaseModel):
    """Base class for :term:`ROIs <roi>` extracted from raw MS data.

    Roi inherits from pydantic BaseModel, and support most of its functionality.
    New ROIs subclasses are created by inheritance of this class and setting
    data fields using Pydantic's standard approach.

    For Numpy array fields, check out the `tidyms.utils.numpy.FloatingArray` and
    `tidyms.utils.numpy.IntArray` types which provide type checking for arrays
    and efficient serialization/deserialization.

    The id field contains a unique identifier for the ROI and is managed internally by
    the library. It **MUST** not be set directly by the user.

    Refer to the :ref:`developer guides <extending-guide>` for an example on how to create a new
    ROI class.

    """

    sample: Sample = pydantic.Field(repr=False)
    """The sample where the ROI was extracted from."""

    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    @classmethod
    def from_str(cls, ser: str, sample: Sample) -> Self:
        """Create a ROI instance from a JSON string.

        :param ser: a serialized ROI obtained using the `to_str` method
        :param sample: a sample to associate with the ROI
        :return: a new ROI instance.

        """
        return cls(sample=sample, **json.loads(ser))

    def to_str(self) -> str:
        """Serialize a ROI into a string.

        :return: a string serialization of the ROI.

        """
        return self.model_dump_json(exclude={"sample"})


class Annotation(TidyMSBaseModel):
    """Store feature annotation data."""

    roi_id: UUID
    """The ROI id where the feature was extracted from"""

    sample_id: str = ""
    """The sample id where the feature was extracted from"""

    group: int = -1
    """The :term:`feature group` id. Group features from different samples based on their chemical
    identity. Used to create a data matrix. If set to ``-1`` the feature is not assigned to any group.
    """

    isotopologue_label: int = -1
    """Group features from the same isotopic envelope in a sample. If set to ``-1`` the feature is
    not associated with any group of isotopologues in a sample."""

    isotopologue_index: int = -1
    """Position of the feature in an isotopic envelope.  If set to ``-1`` the feature is
    not associated with any group of isotopologues in a sample."""

    charge: int = -1
    """Feature charge state. If set to ``-1`` the feature charge state is not defined"""


RoiType = TypeVar("RoiType", bound=Roi)


class Feature(TidyMSBaseModel, Generic[RoiType]):
    """Base class to represent a feature extracted from a ROI.

    Feature inherits from pydantic BaseModel, and support most of its functionality. New Feature
    subclasses are created by inheritance of this class and setting data fields using Pydantic's
    standard approach.

    There are two field types for features:

    Data fields
        contain information to represent the feature. e.g. the start and end position of a
        chromatographic peak. These fields are represented as standard pydantic fields.
    Descriptors
        describe feature characteristics. e.g, the peak width or peak area in a chromatographic
        peak. **ALL** descriptors **MUST** be floats. These fields are represented as pydantic
        computed fields. Descriptors **MUST** be decorated with `pydantic.computed_field`. It is also
        recommended to use the `functools.cached_property` decorator to cache the descriptor value.
        As an example:

        .. code-block:: python

            from functools import cached_property
            from pydantic import computed_field

            class MyFeature(Feature):

                data_field: float = 1.0
                '''A feature data field.'''

                @computed_field
                @cached_property
                def custom_descriptor(self):
                    self.custom_descriptor = 100.0

        The mz, area and height descriptors are set as abstract methods and need
        to be implemented for all concrete Feature classes.

    Finally, three attributes are defined for the Feature class: `id`, `roi` and `annotation`.
    These parameters are managed internally by the library and they MUST never be set directly
    by the user.

    Refer to the :ref:`developer guides <extending-guide>` for an example on how to create a new
    Feature class.

    """

    roi: RoiType = pydantic.Field(repr=False)
    """The ROI where the feature was detected."""

    annotation: Annotation | None = pydantic.Field(repr=False, default=None)
    """Annotation data of the feature."""

    # TODO: remove order functions

    def __lt__(self, other: Self):
        return self.mz < other.mz

    def __le__(self, other: Self):
        return self.mz <= other.mz

    def __gt__(self, other: Self):
        return self.mz > other.mz

    def __ge__(self, other: Self):
        return self.mz >= other.mz

    def __hash__(self) -> int:
        assert self.annotation is not None
        return self.annotation.id.__hash__()

    @cached_property
    @abstractmethod
    def mz(self) -> float:
        """The feature m/z."""
        ...

    @cached_property
    @abstractmethod
    def area(self) -> float:
        """The feature area."""
        ...

    @cached_property
    @abstractmethod
    def height(self) -> float:
        """The feature height."""
        ...

    def has_descriptors_in_range(self, **bounds: tuple[float, float]) -> bool:
        """Check if feature descriptors fall between lower and upper bounds.

        :param bounds: descriptor lower and upper bound values.
        :return: ``True`` if all descriptors fall between the bounds. ``False`` otherwise.

        """
        for name, (lower, upper) in bounds.items():
            if not lower <= self.get(name) <= upper:
                return False
        return True

    def describe(self) -> dict[str, float]:
        """Compute all available descriptors for the feature.

        A descriptor is any method that starts with `get_`.

        :return: a dictionary that maps descriptor names to descriptor values.

        """
        descriptors = dict()
        for descriptor in self.descriptor_names():
            descriptors[descriptor] = self.get(descriptor)
        return descriptors

    def to_str(self) -> str:
        """Serialize the feature data into a string.

        :return: a string serialization of the feature.

        """
        exclude = self.descriptor_names() | {"roi", "annotation"}
        return self.model_dump_json(exclude=exclude)

    @classmethod
    def from_str(cls, s: str, roi: RoiType, annotation: Annotation) -> Self:
        """Create a feature instance from a string.

        :param s: feature string generated with `to_str`.
        :param roi: ROI where the feature was detected.
        :param annotation: the feature annotation.
        :return: a new feature instance.

        """
        d = json.loads(s)
        return cls(roi=roi, annotation=annotation, **d)

    def get(self, descriptor: str) -> float:
        """Compute a descriptor value.

        :param descriptor: the descriptor name.
        :return: the descriptor value.
        :raises ValueError: if an invalid descriptor name is passed.

        """
        try:
            return getattr(self, descriptor)
        except AttributeError as e:
            msg = f"{descriptor} is not a valid descriptor."
            raise ValueError(msg) from e

    @classmethod
    @lru_cache
    def descriptor_names(cls) -> set[str]:
        """Retrieve the available descriptor names.

        :return: the descriptor names.

        """
        return set(cls.model_computed_fields)

    @pydantic.model_validator(mode="after")
    def _add_annotation(self) -> Self:
        if self.annotation is None:
            self.annotation = Annotation(roi_id=self.roi.id, id=self.id, sample_id=self.roi.sample.id)
        return self


class FeatureGroup(pydantic.BaseModel):
    """Store feature group information."""

    group: int = -1
    """the the group id"""

    annotation: GroupAnnotation
    """The feature group annotation."""

    descriptors: dict[str, float]
    """Aggregated feature descriptors."""

    @pydantic.computed_field(repr=False)
    @cached_property
    def mz(self) -> float:
        """The feature m/z."""
        return self.descriptors["mz"]

    def has_descriptors_in_range(self, **filters: tuple[float, float]) -> bool:
        """Check if feature descriptors fall between lower and upper bounds.

        :param bounds: descriptor lower and upper bound values.
        :return: ``True`` if all descriptors fall between the bounds. ``False`` otherwise.

        """
        for name, (lower, upper) in filters.items():
            if name not in self.descriptors:
                msg = f"Feature group {self.group} does not have descriptor {name}."
                raise exceptions.InvalidFeatureDescriptor(msg)
            if not lower <= self.descriptors[name] <= upper:
                return False
        return True


class AnnotableFeature(Feature[RoiType], ABC):
    """Abstract feature class which inherits from Feature.

    Provides extra functionality to perform feature annotation.
    Base feature with also implements methods for feature annotation.

    """

    @abstractmethod
    def compare(self, other) -> float:
        """Compare the similarity between two features.

        Must be a symmetric function that returns a number between 0.0 and 1.0.

        :param other: feature to compare with.
        :return: the similarity between the feature pair.

        """
        ...

    @staticmethod
    @abstractmethod
    def compute_isotopic_envelope(*features) -> IsotopicEnvelope:
        """Compute the isotopic envelope from a list of isotopologue features.

        :param features: the Collection of features used to compute the envelope.
        :return: The normalized isotopic envelope

        """
        ...


class GroupAnnotation(pydantic.BaseModel):
    """Store annotation of a :term:`feature group`."""

    model_config = pydantic.ConfigDict(validate_assignment=True)

    label: int
    """The feature group label. Identifies features across samples that are originated from the
    same ionic species."""

    isotopologue_index: int | None = pydantic.Field(default=None, repr=False)
    """The position of the feature in an envelope. Only defined if an isotopologue annotation
    algorithm was applied to the dataset."""

    isotopologue_group: int | None = pydantic.Field(default=None, repr=False)
    """Label shared between isotopologue features. Only defined if an isotopologue annotation
    algorithm was applied to the dataset."""

    envelope: IsotopicEnvelope | None = pydantic.Field(default=None, repr=False)
    """The m/z and abundance values of the envelope members. Only defined if an isotopologue
    annotation algorithm was applied to the dataset and `isotopologue_index` is ``0``."""

    charge: int | None = pydantic.Field(default=None, repr=False)
    """The numerical charge of the feature. Only defined if an isotopologue annotation algorithm
    was applied to the dataset."""

    name: str | None = None
    """An optional name for the feature group."""


class IsotopicEnvelope(pydantic.BaseModel):
    """Store m/z and normalized abundance of a set of isotopic :term:`envelope`."""

    mz: list[float]
    """The envelope sorted m/z"""

    p: list[float]
    """The envelope normalized abundance"""


class MSSpectrum(TidyMSBaseModel):
    """Representation of a Mass Spectrum."""

    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    index: int = -1

    mz: FloatArray1D
    """Sorted m/z data"""

    int: FloatArray1D
    """Spectral intensity"""

    ms_level: pydantic.PositiveInt = 1
    """MS level of the current spectrum"""

    time: pydantic.NonNegativeFloat = 0.0
    """Acquisition time of the spectrum"""

    centroid: bool = True
    """Set to ``True`` if the spectrum was converted to centroid mode. ``False`` otherwise."""

    def get_nbytes(self) -> int:
        """Get the number of bytes stored in m/z and intensity arrays."""
        return self.int.nbytes + self.mz.nbytes


class MZTrace(Roi):
    """ROI implementation using m/z traces.

    An m/z trace is a 1D trace containing m/z, time and intensity information across scans.

    """

    mz: FloatArray1D = pydantic.Field(repr=False)
    """m/z in each scan. All values are assumed to be non-negative."""

    spint: FloatArray1D = pydantic.Field(repr=False)
    """intensity in each scan. All values are assumed to be non-negative."""

    time: FloatArray1D = pydantic.Field(repr=False)
    """time in each scan. All values are assumed to be non-negative."""

    scan: IntArray1D = pydantic.Field(repr=False)
    """scan numbers where the ROI is defined. All values are assumed to be non-negative."""
    noise: FloatArray1D | None = pydantic.Field(repr=False, default=None)
    """if provided, represent the noise level at teach time point."""

    baseline: FloatArray1D | None = pydantic.Field(repr=False, default=None)
    """if provided, represent the baseline level at teach time point."""

    @pydantic.model_validator(mode="after")
    def check_noise_all_non_negative(self) -> Self:
        """Validate noise."""
        if self.noise is not None:
            msg = "All noise array elements must be non-negative."
            assert numpy.all(self.noise >= 0.0), msg
        return self

    @pydantic.model_validator(mode="after")
    def check_baseline_lower_or_equal_than_intensity(self) -> Self:
        """Validate baseline."""
        if self.baseline is not None:
            msg = "All baseline elements must be lower than their corresponding trace intensity."
            assert numpy.all(self.spint >= self.baseline), msg
        return self

    def get_slice_height(self, start: int, end: int) -> FloatArray1D:
        """Compute the trace height in a slice.

        :param start: slice start index.
        :param end: slice end index.

        """
        height = self.spint[start:end]
        if self.baseline is not None:
            height = height - self.baseline[start:end]
        return numpy.maximum(height, 0.0)

    def equals(self, other: Self) -> bool:
        """Check if two m/z traces are equal."""
        return (
            numpy.array_equal(self.mz, other.mz)
            and numpy.array_equal(self.time, other.time)
            and numpy.array_equal(self.spint, other.spint)
            and numpy.array_equal(self.scan, other.scan)
            and self.id == other.id
        )


class Chromatogram(TidyMSBaseModel):
    """Chromatogram representation."""

    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    index: int = -1
    """The chromatogram order in a data file"""

    name: str | None = None
    """The chromatogram name in a data file"""

    time: FloatArray1D
    """The time data"""

    int: FloatArray1D
    """The intensity data"""


class DescriptorPatch(pydantic.BaseModel):
    """Store a descriptor patch."""

    id: UUID
    """The feature id to patch."""

    descriptor: str
    """The descriptor to patch."""

    value: float
    """The new descriptor value to apply"""


class AnnotationPatch(pydantic.BaseModel):
    """Store an annotation patch."""

    id: UUID
    """The feature id to patch."""

    field: str
    """The annotation field fo patch."""

    value: int
    """The new value."""


class FillValue(pydantic.BaseModel):
    """Container class that stores values to fill missing entries in the data matrix."""

    sample_id: str
    """The id of the sample to input the missing value."""

    feature_group: int
    """The feature group to input the missing value."""

    value: float
    """The fill value."""


FeatureType = TypeVar("FeatureType", bound=Feature)
AnnotableFeatureType = TypeVar("AnnotableFeatureType", bound=AnnotableFeature)



"""Serializable numpy array types."""

from __future__ import annotations

import base64
import json
from typing import Iterable, Literal, TypeVar

import numpy
from numpy import floating, frombuffer, integer
from numpy.typing import NDArray
from pydantic.functional_serializers import PlainSerializer
from pydantic.functional_validators import BeforeValidator
from typing_extensions import Annotated


def array_to_json_str(arr: NDArray) -> str:
    """Serialize a numpy array as a JSON string.

    :param arr: The numpy array to serialize
    :return: JSON string with the following three fields. `dtype` store the array dtype, `shape` contains the array
        shape and `base64_bytes` stores the array data in base64 format.

    """
    d = {
        "dtype": str(arr.dtype),
        "shape": arr.shape,
        "base64_bytes": base64.b64encode(arr.tobytes()).decode("utf8"),
    }
    return json.dumps(d)


def json_str_to_array(s: str):
    """Decode a string generated with array_to_json_str into a numpy array.

    :param s: A string serialized numpy array
    :return: a new numpy instance

    """
    d = json.loads(s)
    dtype = d["dtype"]
    shape = d["shape"]
    data = base64.b64decode(bytes(d["base64_bytes"], "utf8"))
    return frombuffer(data, dtype=dtype).reshape(shape).copy()


def validate_serializable_array(arr: NDArray) -> NDArray:
    """Create an array if a serialized string is provided."""
    if isinstance(arr, str):
        arr = json_str_to_array(arr)
    return arr


def cartesian_product_from_ranges(*args: list[int]) -> IntArray:
    """Create an array where each row is an element in the cartesian product.

    :param args: the list of integers to build the cartesian product

    """
    product = numpy.zeros(shape=(1, 0), dtype=int)

    for x in args:
        n_row, n_col = product.shape
        new_shape = (n_row * len(x), n_col + 1)
        new_product = numpy.zeros(shape=new_shape, dtype=int)
        ind = numpy.repeat(numpy.arange(n_row), len(x))
        new_column = numpy.tile(x, n_row)
        new_product[:, :n_col] = product[ind]
        new_product[:, -1] = new_column
        product = new_product
    return product


def cartesian_product_from_iterable(*args: Iterable):
    """Compute the cartesian product of args as a 2D array."""
    if not args:
        raise ValueError("At least one argument is required to compute the cartesian product.")
    res = None
    for x in args:
        if res is None:
            # initialize cartesian product array
            res = numpy.array(x)
            res = res.reshape((res.size, 1))
        else:
            x = numpy.array(x)
            row, col = res.shape
            new_res_shape = (row * x.size, col + 1)
            new_res = numpy.zeros(shape=new_res_shape, dtype=res.dtype)
            ind = numpy.repeat(numpy.arange(row), x.size)
            new_col = numpy.tile(x, row)
            new_res[:, :col] = res[ind]
            new_res[:, -1] = new_col
            res = new_res
    return res


def _find_closest_sorted(x: numpy.ndarray, xq: numpy.ndarray | float | int) -> IntArray1D:
    """Find the index in x closest to each xq element. Assumes that x is sorted.

    :param x: Sorted 1D array
    :param xq: search vector
    :return: an int array with indices of elements in `x` closest to each `xq` element.
    :raises ValueError: when x or xq are empty.

    """
    if isinstance(xq, (float, int)):
        xq = numpy.array(xq)

    if x.size == 0:
        msg = "`x` must be a non-empty array."
        raise ValueError(msg)

    if xq.size == 0:
        return numpy.array([], dtype=int)

    ind = numpy.searchsorted(x, xq)

    if ind.size == 1:
        if ind == 0:
            return ind
        elif ind == x.size:
            return ind - 1
        else:
            return ind - ((xq - x[ind - 1]) < (x[ind] - xq))

    else:
        # cases where the index is between 1 and x.size - 1
        mask = (ind > 0) & (ind < x.size)
        ind[mask] -= (xq[mask] - x[ind[mask] - 1]) < (x[ind[mask]] - xq[mask])
        # when the index is x.size, then the closest index is x.size -1
        ind[ind == x.size] = x.size - 1
        return ind


def find_closest(x: numpy.ndarray, xq: numpy.ndarray | float | int, is_sorted: bool = True) -> IntArray1D:
    """Find the index in `x` closest to each `xq` element.

    :param x: reference array
    :param xq: search vector
    :param is_sorted: if set to ``True``, assumes that `x` is sorted.
    :return: an int array with indices of elements in `x` closest to each `xq` element.
    :raises ValueError: when x or xq are empty.

    """
    if is_sorted:
        return _find_closest_sorted(x, xq)
    else:
        sorted_index = numpy.argsort(x)
        closest_index = _find_closest_sorted(x[sorted_index], xq)
        return sorted_index[closest_index]


def gauss(grid: numpy.ndarray, mu: float, sigma: float, amp: float) -> numpy.ndarray:
    """Create a gaussian curve from a grid.

    :param grid: an array with grid points used to create the gaussian curve
    :param mu : the gaussian mean
    :param sigma: the gaussian standard deviation
    :param amp: the gaussian amplitude

    """
    gaussian = amp * numpy.power(numpy.e, -0.5 * ((grid - mu) / sigma) ** 2)
    return gaussian


def gaussian_mixture(x: numpy.ndarray, *gaussian_params: tuple[float, float, float]) -> numpy.ndarray:
    """Create a gaussian mixture using a grid.

    :param grid: an array with grid points used to create the gaussian curve
    :param gaussian_params: a tuple of mean, standard deviation and amplitude for each gaussian component
    """
    components = numpy.zeros((len(gaussian_params), x.size))
    for row, component_params in enumerate(gaussian_params):
        components[row] = gauss(x, *component_params)
    return components.sum(axis=0)


def check_matrix_shape(X: NDArray) -> None:
    """Check that the array has the expected data matrix shape."""
    if len(X.shape) != 2:
        raise ValueError("X must be a 2D array.")

    n_rows, n_cols = X.shape

    if n_rows == 0 or n_cols == 0:
        raise ValueError(f"X must be a 2D non-empty array. Got shape {X.shape}")


FloatDtype = TypeVar("FloatDtype", bound=floating)
IntDtype = TypeVar("IntDtype", bound=integer)


FloatArray = Annotated[
    NDArray[FloatDtype],
    BeforeValidator(validate_serializable_array),
    PlainSerializer(array_to_json_str, return_type=str),
]

IntArray = Annotated[
    NDArray[IntDtype],
    BeforeValidator(validate_serializable_array),
    PlainSerializer(array_to_json_str, return_type=str),
]

FloatArray1D = Annotated[
    NDArray[FloatDtype],
    Literal["N"],
    BeforeValidator(validate_serializable_array),
    PlainSerializer(array_to_json_str, return_type=str),
]

IntArray1D = Annotated[
    NDArray[IntDtype],
    Literal["N"],
    BeforeValidator(validate_serializable_array),
    PlainSerializer(array_to_json_str, return_type=str),
]



from typing import cast

import numpy as np
import pytest
from scipy.stats import ConstantInputWarning, median_abs_deviation

from tidyms2.core.enums import CorrelationMethod
from tidyms2.core.utils import metrics
from tidyms2.core.utils.numpy import FloatArray


class TestCV:
    def test_non_2d_array_raise_error(self):
        X = np.arange(5)
        with pytest.raises(ValueError):
            metrics.cv(X)

    def test_2d_empty_arrays_raise_error(self):
        X = np.random.normal(size=(2, 0))
        with pytest.raises(ValueError):
            metrics.cv(X)

    def test_2d_array_single_column(self):
        X = np.array([[1], [2], [3]])
        actual = metrics.cv(X, robust=False)
        expected = np.array([np.std(X, ddof=1)]) / np.array([np.mean(X)])

        assert actual.shape == (1,)
        assert np.allclose(actual, expected)

    def test_single_column_single_value_returns_nan(self):
        X = np.array([[1]])
        with pytest.warns(RuntimeWarning):
            actual = metrics.cv(X, robust=False)

        assert actual.shape == (1,)
        assert np.all(np.isnan(actual))

    def test_single_zeros_column_return_nan(self):
        X = np.zeros(shape=(3, 1))
        with pytest.warns(RuntimeWarning):
            actual = metrics.cv(X, robust=False)

        assert actual.shape == (1,)
        assert np.all(np.isnan(actual))

    def test_single_column_constant_value_return_zero(self):
        X = np.ones(shape=(3, 1))
        actual = metrics.cv(X, robust=False)

        assert actual.shape == (1,)
        assert np.allclose(actual, 0.0)

    def test_single_column_two_non_nans_ok(self):
        X = np.array([[1.0], [np.nan], [1.0]])
        actual = metrics.cv(X, robust=False)

        assert actual.shape == (1,)
        assert np.allclose(actual, 0.0)

    def test_single_column_one_non_nan_return_nan(self):
        X = np.array([[1.0], [np.nan], [np.nan]])
        with pytest.warns(RuntimeWarning):
            actual = metrics.cv(X, robust=False)

        assert actual.shape == (1,)
        assert np.all(np.isnan(actual))

    def test_single_column_all_nans_return_nan(self):
        X = np.array([[np.nan], [np.nan], [np.nan]])
        with pytest.warns(RuntimeWarning):
            actual = metrics.cv(X, robust=False)

        assert actual.shape == (1,)
        assert np.all(np.isnan(actual))

    def test_multiple_column_ok(self):
        _, n_cols = shape = (4, 5)
        X = np.ones(shape=shape)
        actual = metrics.cv(X, robust=False)

        assert actual.shape == (n_cols,)
        assert np.allclose(actual, 0.0)


class TestCVRobust:
    def test_non_2d_array_raise_error(self):
        X = np.arange(5)
        with pytest.raises(ValueError):
            metrics.cv(X, robust=True)

    def test_2d_empty_arrays_raise_error(self):
        X = np.random.normal(size=(2, 0))
        with pytest.raises(ValueError):
            metrics.cv(X, robust=True)

    def test_2d_array_single_column(self):
        X = np.array([[1], [2], [3]])
        actual = metrics.cv(X, robust=True)
        s = median_abs_deviation(X, scale="normal")  # type: ignore
        expected = np.array([s]) / np.array([np.median(X)])

        assert actual.shape == (1,)
        assert np.allclose(actual, expected)

    def test_2d_array_single_column_single_value_returns_zero(self):
        X = np.array([[1.0]])
        actual = metrics.cv(X, robust=True)
        expected = np.array([0.0])

        assert actual.shape == (1,)
        assert np.allclose(actual, expected)

    def test_single_zeros_column_return_nan(self):
        X = np.zeros(shape=(3, 1))
        with pytest.warns(RuntimeWarning):
            actual = metrics.cv(X, robust=True)

        assert actual.shape == (1,)
        assert np.all(np.isnan(actual))

    def test_single_column_constant_value_return_zero(self):
        X = np.ones(shape=(3, 1))
        actual = metrics.cv(X, robust=True)

        assert actual.shape == (1,)
        assert np.allclose(actual, 0.0)

    def test_single_column_with_nans_ignore_nans(self):
        X = np.array([[1.0], [np.nan], [1.0]])
        actual = metrics.cv(X, robust=True)

        assert actual.shape == (1,)
        assert np.allclose(actual, 0.0)

    def test_single_column_all_nans_return_nan(self):
        X = np.array([[np.nan], [np.nan], [np.nan]])
        with pytest.warns(RuntimeWarning):
            actual = metrics.cv(X, robust=True)

        assert actual.shape == (1,)
        assert np.all(np.isnan(actual))

    def test_multiple_column_ok(self):
        _, n_cols = shape = (4, 5)
        X = np.ones(shape=shape)
        actual = metrics.cv(X, robust=True)

        assert actual.shape == (n_cols,)
        assert np.allclose(actual, 0.0)


class TestDetectionRate:
    def test_non_2D_array_raises_error(self):
        X = np.arange(5)
        with pytest.raises(ValueError):
            metrics.detection_rate(X)

    def test_2d_empty_arrays_raise_error(self):
        X = np.random.normal(size=(2, 0))
        with pytest.raises(ValueError):
            metrics.detection_rate(X)

    def test_single_column(self):
        X = np.ones(shape=(3, 1), dtype=float)
        actual = metrics.detection_rate(X)
        expected = np.ones(shape=(1,))

        assert actual.shape == (1,)
        assert np.allclose(actual, expected)

    def test_single_column_nans_are_counted_as_undetected(self):
        X = np.array([[1], [np.nan], [2], [3]], dtype=float)
        actual = metrics.detection_rate(X)
        expected = np.array([0.75])

        assert actual.shape == (1,)
        assert np.allclose(actual, expected)

    def test_single_column_values_below_threshold_are_counted_as_undetected(self):
        X = np.array([[1], [np.nan], [2], [3]], dtype=float)
        actual = metrics.detection_rate(X, threshold=1.5)
        expected = np.array([0.5])

        assert actual.shape == (1,)
        assert np.allclose(actual, expected)

    def test_single_column_values_below_equal_to_threshold_are_counted_as_detected(self):
        X = np.array([[1], [np.nan], [2], [3]], dtype=float)
        actual = metrics.detection_rate(X, threshold=1.0)
        expected = np.array([0.75])

        assert actual.shape == (1,)
        assert np.allclose(actual, expected)

    def test_single_column_nan_threshold_returns_zero(self):
        X = np.array([[1], [np.nan], [2], [3]], dtype=float)
        actual = metrics.detection_rate(X, threshold=np.nan)
        expected = np.array([0.0])

        assert actual.shape == (1,)
        assert np.allclose(actual, expected)

    def test_multiple_columns_scalar_threshold(self):
        X = np.array([[1.0, 2.0], [np.nan, 2.0], [2.0, 2.0], [3.0, 2.0]], dtype=float)
        actual = metrics.detection_rate(X, threshold=1.0)
        expected = np.array([0.75, 1.0])

        assert actual.shape == (2,)
        assert np.allclose(actual, expected)

    def test_multiple_columns_array_threshold(self):
        X = np.array([[1.0, 2.0], [np.nan, 2.0], [2.0, 2.0], [3.0, 2.0]], dtype=float)
        threshold = np.array([1.0, np.nan])
        actual = metrics.detection_rate(X, threshold=threshold)
        expected = np.array([0.75, 0.0])

        assert actual.shape == (2,)
        assert np.allclose(actual, expected)


class TestDRatio:
    def test_non_2d_x_sample_raise_error(self):
        Xs = np.arange(5)
        Xqc = np.ones(shape=(2, 2), dtype=float)
        with pytest.raises(ValueError):
            metrics.dratio(Xs, Xqc)

    def test_non_2d_x_qc_raise_error(self):
        Xs = np.ones(shape=(2, 2), dtype=float)
        Xqc = np.arange(5)
        with pytest.raises(ValueError):
            metrics.dratio(Xs, Xqc)

    def test_2d_empty_x_sample_raise_error(self):
        Xs = np.random.normal(size=(2, 0))
        Xqc = np.ones(shape=(2, 2), dtype=float)
        with pytest.raises(ValueError):
            metrics.dratio(Xs, Xqc)

    def test_2d_empty_x_qc_raise_error(self):
        Xs = np.ones(shape=(2, 2), dtype=float)
        Xqc = np.random.normal(size=(2, 0))
        with pytest.raises(ValueError):
            metrics.dratio(Xs, Xqc)

    def test_single_column_single_row_x_sample_return_nan(self):
        Xs = np.array([[1]])
        Xqc = np.array([[1], [2], [3]])
        with pytest.warns(RuntimeWarning):
            actual = metrics.dratio(Xs, Xqc, robust=False)

        assert actual.shape == (1,)
        assert np.all(np.isnan(actual))

    def test_single_column_single_row_x_qc_return_nan(self):
        Xs = np.array([[1], [2], [3]])
        Xqc = np.array([[1]])
        with pytest.warns(RuntimeWarning):
            actual = metrics.dratio(Xs, Xqc, robust=False)

        assert actual.shape == (1,)
        assert np.all(np.isnan(actual))

    def test_single_column_one_non_nan_in_x_sample_return_nan(self):
        Xs = np.array([[1], [np.nan], [np.nan]])
        Xqc = np.array([[1], [np.nan], [3]])
        with pytest.warns(RuntimeWarning):
            actual = metrics.dratio(Xs, Xqc, robust=False)

        assert actual.shape == (1,)
        assert np.all(np.isnan(actual))

    def test_single_column_one_non_nan_in_x_qc_return_nan(self):
        Xs = np.array([[1], [np.nan], [3]])
        Xqc = np.array([[1], [np.nan], [np.nan]])
        with pytest.warns(RuntimeWarning):
            actual = metrics.dratio(Xs, Xqc, robust=False)

        assert actual.shape == (1,)
        assert np.all(np.isnan(actual))

    def test_single_column_two_non_nans_in_x_sample_ok(self):
        Xs = np.array([[1], [np.nan], [3]])
        Xqc = np.array([[1], [np.nan], [3]])
        actual = metrics.dratio(Xs, Xqc, robust=False)
        expected = np.array([1.0])

        assert actual.shape == (1,)
        assert np.allclose(actual, expected)

    def test_constant_values_in_x_qc_returns_zero(self):
        Xs = np.array([[1], [2.0], [3]])
        Xqc = np.array([[1.0], [1.0], [1.0]])
        actual = metrics.dratio(Xs, Xqc, robust=False)
        expected = np.array([0.0])

        assert actual.shape == (1,)
        assert np.allclose(actual, expected)

    def test_constant_values_in_x_sample_returns_nan(self):
        Xs = np.array([[1.0], [1.0], [1.0]])
        Xqc = np.array([[1], [2.0], [3]])
        with pytest.warns(RuntimeWarning):
            actual = metrics.dratio(Xs, Xqc, robust=False)

        assert actual.shape == (1,)
        assert np.all(np.isinf(actual))

    def test_multiple_columns_ok(self):
        Xs = np.array([[1.0, 1.0], [2.0, 2.0], [3.0, 3.0]])
        Xqc = np.array([[1.0, 1.0], [2.0, 1.0], [3.0, 1.0]])
        actual = metrics.dratio(Xs, Xqc, robust=False)
        expected = np.array([1.0, 0.0])

        assert actual.shape == (2,)
        assert np.allclose(actual, expected)


class TestDRatioRobust:
    def test_non_2d_x_sample_raise_error(self):
        Xs = np.arange(5)
        Xqc = np.ones(shape=(2, 2), dtype=float)
        with pytest.raises(ValueError):
            metrics.dratio(Xs, Xqc, robust=True)

    def test_non_2d_x_qc_raise_error(self):
        Xs = np.ones(shape=(2, 2), dtype=float)
        Xqc = np.arange(5)
        with pytest.raises(ValueError):
            metrics.dratio(Xs, Xqc, robust=True)

    def test_2d_empty_x_sample_raise_error(self):
        Xs = np.random.normal(size=(2, 0))
        Xqc = np.ones(shape=(2, 2), dtype=float)
        with pytest.raises(ValueError):
            metrics.dratio(Xs, Xqc, robust=True)

    def test_2d_empty_x_qc_raise_error(self):
        Xs = np.ones(shape=(2, 2), dtype=float)
        Xqc = np.random.normal(size=(2, 0))
        with pytest.raises(ValueError):
            metrics.dratio(Xs, Xqc, robust=True)

    def test_single_column_single_row_x_sample_return_inf(self):
        Xs = np.array([[1]])
        Xqc = np.array([[1], [2], [3]])
        with pytest.warns(RuntimeWarning):
            actual = metrics.dratio(Xs, Xqc, robust=True)

        assert actual.shape == (1,)
        assert np.all(np.isinf(actual))

    def test_single_column_single_row_x_qc_return_nan(self):
        Xs = np.array([[1], [2], [3]])
        Xqc = np.array([[1]])
        actual = metrics.dratio(Xs, Xqc, robust=True)

        expected = np.zeros(shape=(1,), dtype=float)

        assert actual.shape == (1,)
        assert np.allclose(actual, expected)

    def test_single_column_one_non_nan_in_x_sample_ok(self):
        Xs = np.array([[1], [np.nan], [np.nan]])
        Xqc = np.array([[1], [np.nan], [3]])
        with pytest.warns(RuntimeWarning):
            actual = metrics.dratio(Xs, Xqc, robust=True)

        assert actual.shape == (1,)
        assert np.all(np.isinf(actual))

    def test_single_column_one_non_nan_in_x_qc_return_zero(self):
        Xs = np.array([[1], [np.nan], [3]])
        Xqc = np.array([[1], [np.nan], [np.nan]])
        actual = metrics.dratio(Xs, Xqc, robust=True)

        expected = np.zeros(shape=(1,), dtype=float)
        assert actual.shape == (1,)
        assert np.allclose(actual, expected)

    def test_single_column_two_non_nans_in_x_sample_ok(self):
        Xs = np.array([[1], [np.nan], [3]])
        Xqc = np.array([[1], [np.nan], [3]])
        actual = metrics.dratio(Xs, Xqc, robust=True)
        expected = np.array([1.0])

        assert actual.shape == (1,)
        assert np.allclose(actual, expected)

    def test_constant_values_in_x_qc_returns_zero(self):
        Xs = np.array([[1], [2.0], [3]])
        Xqc = np.array([[1.0], [1.0], [1.0]])
        actual = metrics.dratio(Xs, Xqc, robust=True)
        expected = np.array([0.0])

        assert actual.shape == (1,)
        assert np.allclose(actual, expected)

    def test_constant_values_in_x_sample_returns_nan(self):
        Xs = np.array([[1.0], [1.0], [1.0]])
        Xqc = np.array([[1], [2.0], [3]])
        with pytest.warns(RuntimeWarning):
            actual = metrics.dratio(Xs, Xqc, robust=True)

        assert actual.shape == (1,)
        assert np.all(np.isinf(actual))

    def test_multiple_columns_ok(self):
        Xs = np.array([[1.0, 1.0], [2.0, 2.0], [3.0, 3.0]])
        Xqc = np.array([[1.0, 1.0], [2.0, 1.0], [3.0, 1.0]])
        actual = metrics.dratio(Xs, Xqc, robust=True)
        expected = np.array([1.0, 0.0])

        assert actual.shape == (2,)
        assert np.allclose(actual, expected)


class TestPCA:
    def test_return_only_scores(self):
        n_rows = 20
        n_cols = 10
        n_components = 2
        X = cast(FloatArray, np.random.normal(size=(n_rows, n_cols)))
        scores, loadings, variance = metrics.pca(X)
        assert scores.shape == (n_rows, n_components)
        assert loadings is None
        assert variance is None

    def test_return_loadings(self):
        n_rows = 20
        n_cols = 10
        n_components = 2
        X = np.random.normal(size=(n_rows, n_cols))
        scores, loadings, variance = metrics.pca(X, n_components=n_components, return_loadings=True)
        assert scores.shape == (n_rows, n_components)
        assert loadings is not None
        assert loadings.shape == (n_cols, n_components)
        assert variance is None

    def test_return_variance(self):
        n_rows = 20
        n_cols = 10
        n_components = 2

        X = np.random.normal(size=(n_rows, n_cols))
        scores, loadings, variance = metrics.pca(X, n_components=n_components, return_variance=True)

        assert scores.shape == (n_rows, n_components)
        assert loadings is None
        assert variance is not None
        assert variance.shape == (n_components,)

    def test_return_variance_and_return_loadings(self):
        n_rows = 20
        n_cols = 10
        n_components = 2
        X = np.random.normal(size=(n_rows, n_cols))

        scores, loadings, variance = metrics.pca(
            X,
            n_components=n_components,
            return_loadings=True,
            return_variance=True,
        )
        assert scores.shape == (n_rows, n_components)
        assert loadings is not None
        assert loadings.shape == (n_cols, n_components)
        assert variance is not None
        assert variance.shape == (n_components,)

    def test_with_scaling(self):
        n_rows = 20
        n_cols = 10
        n_components = 2
        X = np.random.normal(size=(n_rows, n_cols))

        scores, loadings, variance = metrics.pca(
            X,
            n_components=n_components,
            scaling="autoscaling",
            return_loadings=True,
            return_variance=True,
        )
        assert scores.shape == (n_rows, n_components)
        assert loadings is not None
        assert loadings.shape == (n_cols, n_components)
        assert variance is not None
        assert variance.shape == (n_components,)
        assert ~np.any(np.isnan(scores))

    def test_with_normalization(self):
        n_rows = 20
        n_cols = 10
        n_components = 2
        X = np.random.normal(size=(n_rows, n_cols))

        scores, loadings, variance = metrics.pca(
            X,
            n_components=n_components,
            normalization="sum",
            return_loadings=True,
            return_variance=True,
        )
        assert scores.shape == (n_rows, n_components)
        assert loadings is not None
        assert loadings.shape == (n_cols, n_components)
        assert variance is not None
        assert variance.shape == (n_components,)
        assert ~np.any(np.isnan(scores))

    def test_with_normalization_and_scaling(self):
        n_rows = 20
        n_cols = 10
        n_components = 2
        X = np.random.normal(size=(n_rows, n_cols))

        scores, loadings, variance = metrics.pca(
            X,
            n_components=n_components,
            normalization="sum",
            scaling="autoscaling",
            return_loadings=True,
            return_variance=True,
        )
        assert scores.shape == (n_rows, n_components)
        assert loadings is not None
        assert loadings.shape == (n_cols, n_components)
        assert variance is not None
        assert variance.shape == (n_components,)
        assert ~np.any(np.isnan(scores))

    def test_n_components_greater_than_rank_raises_error(self):
        n_rows = 20
        n_cols = 10
        X = np.random.normal(size=(n_rows, n_cols))

        with pytest.raises(ValueError):
            metrics.pca(X, n_components=n_rows)

    def test_matrix_with_nan_raises_error(self):
        n_rows = 20
        n_cols = 10
        n_components = 2
        X = np.random.normal(size=(n_rows, n_cols))
        X[1:1] = np.nan
        X[2:4] = np.nan

        with pytest.raises(ValueError):
            metrics.pca(X, n_components=n_components)


class TestCorrelationPearson:
    def test_single_row_raises_error(self):
        y = np.array([1.0])
        x = np.array([[1.0]])

        with pytest.raises(ValueError):
            metrics.correlation(x, y, method=CorrelationMethod.PEARSON)

    def test_non_matching_shapes_raise_error(self):
        y = np.array([1.0])
        x = np.array([[1.0, 2.0]])

        with pytest.raises(ValueError):
            metrics.correlation(x, y, method=CorrelationMethod.PEARSON)

    def test_single_column(self):
        y = np.array([1.0, 2.0])
        x = np.array([[1.0], [2.0]])

        actual = metrics.correlation(x, y, method=CorrelationMethod.PEARSON)
        expected = np.array([1.0])
        assert np.allclose(actual, expected)

    def test_single_column_with_constant_values_return_nan(self):
        y = np.array([1.0, 1.0])
        x = np.array([[1.0], [1.0]])

        with pytest.warns(ConstantInputWarning):
            actual = metrics.correlation(x, y, method=CorrelationMethod.PEARSON)
        assert actual.shape == (1,)
        assert np.all(np.isnan(actual))

    def test_single_column_with_nan_returns_nan(self):
        y = np.array([1.0, 2.0])
        x = np.array([[1.0], [np.nan]])

        actual = metrics.correlation(x, y, method=CorrelationMethod.PEARSON)
        assert actual.shape == (1,)
        assert np.all(np.isnan(actual))

    def test_multiple_columns(self):
        y = np.array([1.0, 2.0])
        x = np.array([[1.0, -1.0, 1.0], [-1.0, 2.0, np.nan]])

        actual = metrics.correlation(x, y, method=CorrelationMethod.PEARSON)
        expected = np.array([-1.0, 1.0, np.nan])
        assert actual.shape == (3,)
        assert np.allclose(actual, expected, equal_nan=True)


class TestCorrelationSpearman:
    def test_single_row_raises_error(self):
        y = np.array([1.0])
        x = np.array([[1.0]])

        with pytest.raises(ValueError):
            metrics.correlation(x, y, method=CorrelationMethod.SPEARMAN)

    def test_non_matching_shapes_raise_error(self):
        y = np.array([1.0])
        x = np.array([[1.0, 2.0]])

        with pytest.raises(ValueError):
            metrics.correlation(x, y, method=CorrelationMethod.SPEARMAN)

    def test_single_column(self):
        y = np.array([1.0, 2.0])
        x = np.array([[1.0], [2.0]])

        actual = metrics.correlation(x, y, method=CorrelationMethod.SPEARMAN)
        expected = np.array([1.0])
        assert np.allclose(actual, expected)

    def test_single_column_with_constant_values_return_nan(self):
        y = np.array([1.0, 1.0])
        x = np.array([[1.0], [1.0]])

        with pytest.warns(ConstantInputWarning):
            actual = metrics.correlation(x, y, method=CorrelationMethod.SPEARMAN)
        assert actual.shape == (1,)
        assert np.all(np.isnan(actual))

    def test_single_column_with_nan_returns_nan(self):
        y = np.array([1.0, 2.0])
        x = np.array([[1.0], [np.nan]])

        actual = metrics.correlation(x, y, method=CorrelationMethod.SPEARMAN)
        assert actual.shape == (1,)
        assert np.all(np.isnan(actual))

    def test_multiple_columns(self):
        y = np.array([1.0, 2.0])
        x = np.array([[1.0, -1.0, 1.0], [-1.0, 2.0, np.nan]])

        actual = metrics.correlation(x, y, method=CorrelationMethod.SPEARMAN)
        expected = np.array([-1.0, 1.0, np.nan])
        assert actual.shape == (3,)
        assert np.allclose(actual, expected, equal_nan=True)



"""Test lcms and fileio functionality with real data."""

import pathlib

import pytest

from tidyms2.core.models import Chromatogram, MSSpectrum
from tidyms2.io import MSData


@pytest.mark.parametrize(
    "filename",
    [
        "centroid-data-indexed-uncompressed.mzML",
        "centroid-data-zlib-no-index-compressed.mzML",
        "centroid-data-zlib-indexed-compressed.mzML",
        "profile-data-zlib-indexed-compressed.mzML",
    ],
)
class TestMZMLReader:
    def test_get_n_spectra(self, raw_data_dir: pathlib.Path, filename: str):
        ms_data = MSData(raw_data_dir / filename)
        n = ms_data.get_n_spectra()
        assert isinstance(n, int)
        assert n > 0

    def test_get_n_chromatograms(self, raw_data_dir: pathlib.Path, filename: str):
        ms_data = MSData(raw_data_dir / filename)
        n = ms_data.get_n_chromatograms()
        assert isinstance(n, int)
        assert n > 0

    def test_get_spectrum(self, raw_data_dir: pathlib.Path, filename: str):
        ms_data = MSData(raw_data_dir / filename)
        sp = ms_data.get_spectrum(0)
        assert isinstance(sp, MSSpectrum)

    def test_get_all_spectra(self, raw_data_dir: pathlib.Path, filename: str):
        ms_data = MSData(raw_data_dir / filename)
        expected_n_spectra = ms_data.get_n_spectra()
        all_spectra = [ms_data.get_spectrum(x) for x in range(expected_n_spectra)]

        assert len(all_spectra) == expected_n_spectra
        assert all(isinstance(x, MSSpectrum) for x in all_spectra)

    def test_get_all_spectra_rebuild_index(self, raw_data_dir: pathlib.Path, filename: str):
        ms_data = MSData(raw_data_dir / filename, rebuild_index=True)
        expected_n_spectra = ms_data.get_n_spectra()
        all_spectra = [ms_data.get_spectrum(x) for x in range(expected_n_spectra)]

        assert len(all_spectra) == expected_n_spectra
        assert all(isinstance(x, MSSpectrum) for x in all_spectra)

    def test_get_chromatogram(self, raw_data_dir: pathlib.Path, filename: str):
        ms_data = MSData(raw_data_dir / filename)
        chromatogram = ms_data.get_chromatogram(0)
        assert isinstance(chromatogram, Chromatogram)

    def test_get_all_chromatograms(self, raw_data_dir: pathlib.Path, filename: str):
        ms_data = MSData(raw_data_dir / filename)
        expected_n_chromatograms = ms_data.get_n_chromatograms()
        all_spectra = [ms_data.get_chromatogram(x) for x in range(expected_n_chromatograms)]

        assert len(all_spectra) == expected_n_chromatograms
        assert all(isinstance(x, Chromatogram) for x in all_spectra)



"""Utilities to trigger sample pipeline execution."""

import concurrent.futures
from logging import getLogger
from typing import Generic, Protocol

import pydantic

from ..core.exceptions import SampleProcessorError
from ..core.models import FeatureType, RoiType
from ..core.operators.pipeline import Pipeline
from ..core.storage import AssayStorage, SampleStorage, delete_empty_rois

logger = getLogger(__name__)


class SampleProcessor(Protocol, Generic[RoiType, FeatureType]):
    """Base sample executor class."""

    def execute(
        self,
        storage: AssayStorage[RoiType, FeatureType],
        pipe: Pipeline,
        *samples: SampleStorage[RoiType, FeatureType],
    ) -> None:
        """Apply pipeline to multiple samples and store results into an assay storage."""
        ...


class SequentialSampleProcessor(pydantic.BaseModel, Generic[RoiType, FeatureType]):
    """Execute a sample pipeline."""

    remove_empty_roi: bool = True
    """Flag to delete ROIs that do not contain any features."""

    def execute(
        self,
        storage: AssayStorage[RoiType, FeatureType],
        pipe: Pipeline,
        *samples: SampleStorage[RoiType, FeatureType],
    ) -> None:
        """Apply a pipeline to a sample."""
        n_samples = len(samples)
        for k, sample in enumerate(samples, start=1):
            logger.info(f"Processing sample `{sample.get_sample().id}` ({k}/{n_samples}).")
            pipe.apply(sample)

            if self.remove_empty_roi:
                delete_empty_rois(sample)

            storage.add_sample_data(sample)


class ParallelSampleProcessor(pydantic.BaseModel, Generic[RoiType, FeatureType]):
    """Execute a sample pipeline."""

    remove_empty_roi: bool = True
    """Flag to delete ROIs that do not contain any features."""

    max_workers: pydantic.PositiveInt = 2
    """The maximum number of process spawned simultaneously to process samples."""

    def execute(
        self,
        storage: AssayStorage[RoiType, FeatureType],
        pipe: Pipeline,
        *samples: SampleStorage[RoiType, FeatureType],
    ) -> None:
        """Apply a pipeline to multiple samples and store the results in an assay storage."""
        with concurrent.futures.ProcessPoolExecutor(max_workers=self.max_workers) as executor:
            n_samples = len(samples)
            futures = [executor.submit(_sample_executor_worker, pipe.copy(), sample) for sample in samples]
            for k, future in enumerate(concurrent.futures.as_completed(futures), start=1):
                data = future.result()
                logger.info(f"Processed sample `{data.get_sample().id}` ({k}/{n_samples}).")
                if self.remove_empty_roi:
                    delete_empty_rois(data)
                storage.add_sample_data(data)


def _sample_executor_worker(
    pipeline: Pipeline, data: SampleStorage[RoiType, FeatureType]
) -> SampleStorage[RoiType, FeatureType]:
    """Apply pipeline to a sample data instance."""
    try:
        pipeline.apply(data)
        delete_empty_rois(data)
    except Exception as e:
        raise SampleProcessorError(f"Failed to process sample {data.get_sample().id}") from e
    return data



"""Assay configuration."""

from pathlib import Path

import pydantic

from ..core.models import Sample


class AssayConfiguration(pydantic.BaseModel):
    """The assay configuration model."""

    samples: list[Sample]
    """The list of samples included in the assay."""

    sample_root: Path | None = None
    """The base directory to search for sample paths."""

    storage_type: str
    """the name of a registered storage backend."""

    storage_config: dict
    """The configuration passed to the storage backend."""

    sample_processor_type: str
    """the name of a registered sample processor backend."""

    sample_processor_config: dict
    """The configuration passed to the storage processor."""

    sample_pipeline: list[dict]
    """The configuration of sample operators."""

    assay_pipeline: list[dict]
    """The configuration of assay operators."""

    matrix_pipeline: list[dict]
    """The configuration of the matrix operators."""



"""Models with chemistry utilities configuration."""

from typing import Self

import pydantic

from .table import PTABLE


class ChemicalContextConfiguration(pydantic.BaseModel):
    """Store chemical context configuration."""

    custom_abundances: dict[str, dict[int, float]] | None = None
    """Override natural abundances of isotopes. Custom abundances is a dictionary that maps
    atomic numbers to a dictionary of isotopes mass numbers and their abundance. If an abundance
    is not provided for a particular isotope it is assumed to be zero. The sum of abundances for
    a set of isotopologues must be normalized to one."""


class FormulaGeneratorConfiguration(pydantic.BaseModel):
    """Store the formula generator parameters."""

    bounds: dict[str, tuple[pydantic.NonNegativeInt, pydantic.NonNegativeInt]]
    """A dictionary that maps element (eg: ``"C"``) or isotopes (eg: ``"13C"``) symbols to
    minimum and maximum values of formula coefficients in generated formulas. If element
    symbols are provided, the most abundant isotope is used for formula generation."""

    max_M: pydantic.PositiveFloat
    """Maximum mass value for generated formulas."""

    def update_bounds(self, bounds: dict[str, tuple[int, int]]) -> None:
        """Update or add new bounds."""
        self.bounds.update(bounds)

    @classmethod
    def from_chnops(cls, m: int, **kwargs) -> Self:
        """Create a new instance with predefined bounds for CHNOPS elements.

        CHNOPS bounds were computed by finding the minimum and maximum coefficient bounds for all molecules
        in the `HMDB <http://hmdb.ca>`_ under a specific mass threshold. This function offers precomputed
        bounds using all molecules with mass values under 500, 1000, 1500 and 2000.

        :param m: maximum mass of molecules used to build bounds. Valid values are ``500``, ``1000``,
            ``1500`` or ``2000``.
        :param kwargs: extra arguments passed to the constructor.

        """
        if m == 500:
            bounds = {"C": (0, 34), "H": (0, 70), "N": (0, 10), "O": (0, 18), "P": (0, 4), "S": (0, 7)}
        elif m == 1000:
            bounds = {"C": (0, 70), "H": (0, 128), "N": (0, 15), "O": (0, 31), "P": (0, 8), "S": (0, 7)}
        elif m == 1500:
            bounds = {"C": (0, 100), "H": (0, 164), "N": (0, 23), "O": (0, 46), "P": (0, 8), "S": (0, 7)}
        elif m == 2000:
            bounds = {"C": (0, 108), "H": (0, 190), "N": (0, 23), "O": (0, 61), "P": (0, 8), "S": (0, 8)}
        else:
            raise ValueError(f"Valid mass values are 500, 1000, 1500 or 2000. Got {m}.")

        return cls(bounds=bounds, max_M=m, **kwargs)

    @pydantic.model_validator(mode="before")
    @classmethod
    def _validate_bounds(cls, data):
        for isotope, (lower, upper) in data["bounds"].items():
            _ = PTABLE.get_isotope(isotope)
            if lower >= upper:
                raise ValueError(f"Expected lower bound <= upper bound. Got ({lower}, {upper}) for {isotope}.")
        return data


class EnvelopeScorerConfiguration(FormulaGeneratorConfiguration):
    """Store the envelope scorer configuration."""

    max_length: pydantic.PositiveInt = 5
    """The length of the generated envelopes."""

    context: ChemicalContextConfiguration | None = None
    """The chemical context configuration. If set to ``None`` uses the default chemical context."""


class EnvelopeValidatorConfiguration(EnvelopeScorerConfiguration):
    """Store the envelope validator configuration."""

    min_M_tol: pydantic.PositiveFloat = 0.01
    """Exact mass tolerance for high abundance isotopologues. If ``None``, the parameter is set
    based on the `mode` value. See the notes for an explanation of how this value is used."""

    max_M_tol: pydantic.PositiveFloat = 0.01
    """Exact mass tolerance for low abundance isotopologues.  If ``None``, the parameter is set
    based on the `mode` value. See the notes for an explanation of how this value is used."""

    p_tol: pydantic.PositiveFloat = 0.05
    """Tolerance threshold to include in the abundance results"""

    def remove_elements_with_a_single_isotope(self) -> None:
        """Remove elements with a single isotope from the bounds."""
        self.bounds = {k: v for k, v in self.bounds.items() if len(PTABLE.get_element(k).isotopes) > 1}



import numpy as np
import matplotlib.pyplot as plt

# always generate the same plot
np.random.seed(1234)

amp, width, rt = 30.0, 2.0, 25.0
time = np.arange(50)
signal =  amp * np.power(np.e, -0.5 * ((time - rt) / width) ** 2)
noise = np.random.normal(size=signal.size, scale=1)
x = signal + noise + 3

start, apex, end = 19, 25, 30

fig, ax = plt.subplots(figsize=(6, 6))
ax.plot(time, x, label="signal")
ax.scatter(time[start], x[start], label="peak start", s=50)
ax.scatter(time[apex], x[apex], label="peak apex", s=50)
ax.scatter(time[end], x[end], label="peak end", s=50)
ax.fill_between(time[start:end + 1], x[start:end + 1], alpha=0.2, label="peak region")
ax.annotate(
    text='',
    xy=(time[end + 5], x[end]),
    xytext=(time[end + 5], x[apex]),
    arrowprops=dict(arrowstyle='<->')
)
ax.annotate(text='peak \n prominence', xy=(time[end + 10], x[apex] / 2))
ax.legend()



import numpy as np
import pytest
from scipy.ndimage import gaussian_filter1d
from scipy.signal.windows import gaussian
from scipy.special import erfc

from tidyms2.algorithms import signal
from tidyms2.core.utils.numpy import gaussian_mixture

SEED = 1234

SIGMA = 1.0


@pytest.fixture(scope="module")
def noise():
    np.random.seed(SEED)
    return np.random.normal(size=500, scale=SIGMA)


class TestEstimateLocalNoise:
    def test_empty_signal_returns_zero(self):
        x = np.array([])
        noise = signal._estimate_local_noise(x)
        assert np.isclose(noise, 0.0)

    @pytest.mark.parametrize("x", [np.array([1]), np.array([1, 2])])
    def test_signal_length_lower_than_two(self, x):
        noise = signal._estimate_local_noise(x)
        assert np.isclose(noise, 0.0)

    @pytest.mark.parametrize("robust", [False, True])
    def test_noise_estimation_from_gaussian_noise_is_close_to_std(self, noise, robust):
        # check that the noise estimation is close to the std of a normal distribution
        actual = signal._estimate_local_noise(noise, robust=robust)
        # noise should be close to sigma, check with a 20 % tolerance
        assert abs(actual - SIGMA) / SIGMA < 0.2


class TestEstimateNoise:
    def empty_array_returns_empty_array(self):
        x = np.array([])
        noise = signal.estimate_noise(x)
        assert noise.size == 0.0

    @pytest.mark.parametrize("x", [np.array([1]), np.array([1, 3]), np.array([1, 4, 6])])
    def test_signal_length_lower_than_two(self, x):
        noise_estimation = signal.estimate_noise(x)
        assert np.allclose(noise_estimation, 0.0)

    def test_noise_estimation_size(self, noise):
        n_slices = 2
        actual = signal.estimate_noise(noise, n_chunks=n_slices)
        assert np.all(actual >= 0.0)
        assert np.unique(actual).size == n_slices
        assert noise.size == actual.size

    def test_estimate_noise_number_of_unique_values_in_estimation(self, noise):
        n_slices = 2
        actual = signal.estimate_noise(noise, n_chunks=n_slices)
        assert np.all(actual >= 0.0)
        assert np.unique(actual).size == n_slices

    def test_min_slice_size_has_higher_priority_than_n_slices_to_set_slice_size(self, noise):
        n_slices = 5
        min_slice_size = 150
        noise_estimation = signal.estimate_noise(noise, n_chunks=n_slices, min_chunk_size=min_slice_size)
        # noise has a size of 500, the slice is going to be 100 < 150 check that
        # 150 is used instead. There should be three unique values
        actual_n_slices = np.unique(noise_estimation).size
        assert actual_n_slices == 3


class TestFindLocalExtrema:
    def test_first_middle_and_last_indices_are_extrema_in_triangular_pulse(self):
        x = np.arange(10)
        triangular_pulse = np.hstack((x, x[::-1]))
        actual = signal._find_local_extrema(triangular_pulse)
        expected = [0, 9, 19]
        assert np.array_equal(actual, expected)

    def test_monotone_signal_has_no_local_extrema(self):
        monotone_signal = np.arange(10)
        actual = signal._find_local_extrema(monotone_signal)
        expected = np.array([])
        assert np.array_equal(actual, expected)


class TestBaseline:
    test_noise_sum_params = [[np.array([0, 1]), np.sqrt([25, 25])], [np.array([0]), np.sqrt([34])]]

    @pytest.mark.parametrize("index,expected", test_noise_sum_params)
    def test_get_noise_sum_slice_std(self, index, expected):
        index = np.array(index)
        expected = np.array(expected)
        x = np.array([3, 4, 2, 2, 1])
        test_output = signal._get_noise_slice_sum_std(x, index)
        assert np.allclose(test_output, expected)

    def test_estimate_noise_probability(self):
        noise = np.ones(7)
        x = np.array([0, 0.1, 0.4, 2, 1.25, 1.1, 1.0])
        extrema = np.array([0, 3, 6])
        # two slices of size 4 and 2 respectively, the expected output should
        # be erfc(1/sqrt(2) and erfc(1)
        expected_output = erfc([2.5 * np.sqrt(1 / 2) / 2, 1.35 * np.sqrt(1 / 2) / 2])
        test_output = signal._estimate_noise_probability(noise, x, extrema)
        assert np.allclose(expected_output, test_output)

    def test_build_baseline_index(self):
        x = np.array([0, 1, 2, 1, 0, 1, 2, 1, 0, 1, 2, 1, 0])
        extrema = np.array([0, 2, 4, 6, 8, 10, 12])
        noise_probability = np.array([0, 0.25, 0.25, 0.25, 0, 0])
        min_proba = 0.05
        expected = np.array([0, 4, 5, 6, 12])
        test = signal._build_baseline_index(x, noise_probability, min_proba, extrema)
        assert np.array_equal(expected, test)

    def test_estimate_baseline(self):
        # a simple test, a noise array is built using a noise level greater
        # than the noise level in the signal. All points should be classified as
        # baseline
        n = 100
        x = np.random.normal(size=n, scale=1)
        noise = np.ones(n) * 5
        baseline = signal.estimate_baseline(x, noise)
        expected_baseline_index = np.arange(n)
        test_baseline_index = np.where(np.abs(x - baseline) < noise)[0]
        assert np.array_equal(expected_baseline_index, test_baseline_index)


class TestDetectPeaks:
    @pytest.fixture(scope="class")
    def single_peak(self, noise):
        x = gaussian(noise.size, 2) * 20
        return x

    @pytest.fixture(scope="class")
    def two_non_overlapping_peaks(self, noise):
        x = np.arange(noise.size)
        params = [(100, 2, 50), (150, 2, 25)]
        return gaussian_mixture(x, *params)

    @pytest.fixture
    def two_overlapping_peaks(self, noise):
        x = np.arange(noise.size)
        params = ((100, 2, 50), (108, 2, 25))
        return gaussian_mixture(x, *params)

    def test_signal_with_one_peak_detect_single_peak(self, single_peak, noise):
        x = single_peak + noise
        noise_estimation = signal.estimate_noise(x)

        # smooth x to reduce the number of detected peaks
        x = gaussian_filter1d(x, 1.0)

        baseline_estimation = signal.estimate_baseline(x, noise)
        peak_list = signal.detect_peaks(x, noise_estimation, baseline_estimation)

        assert len(peak_list[0]) == 1

    def test_detect_peaks_two_non_overlapping_peaks(self, two_non_overlapping_peaks, noise):
        x = two_non_overlapping_peaks + noise
        noise_estimation = signal.estimate_noise(x)
        # smooth x to reduce the number of detected peaks
        x = gaussian_filter1d(x, 1.0)
        baseline_estimation = signal.estimate_baseline(x, noise)
        peak_list = signal.detect_peaks(x, noise_estimation, baseline_estimation)
        assert len(peak_list[0]) == 2

    def test_detect_peaks_two_overlapping_peaks(self, two_overlapping_peaks, noise):
        x = two_overlapping_peaks + noise
        noise_estimation = signal.estimate_noise(x)
        # smooth x to reduce the number of detected peaks
        x = gaussian_filter1d(x, 1.0)

        baseline_estimation = signal.estimate_baseline(x, noise)

        peak_list = signal.detect_peaks(x, noise_estimation, baseline_estimation)
        start, apex, end = peak_list

        expected_n_peaks = 2
        assert len(start) == expected_n_peaks

        # check the boundary of the overlapping peaks
        assert end[0] == (start[1] + 1)
